"""Data objects related to building, signing, and broadcasting transactions."""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import List, Optional

from jigu.core.event import Event
from jigu.core.msg import MSG_TYPES, StdMsg
from jigu.core.sdk import Coin, Coins, PublicKey, Timestamp
from jigu.query.event import EventsQuery
from jigu.query.msginfo import MsgInfo, MsgInfosQuery
from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S


__all__ = [
    "StdFee",
    "StdSignature",
    "StdTx",
    "StdSignMsg",
    "TxInfo",
    "TxBroadcastResult",
]


@dataclass
class StdFee(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(gas=S.STRING_INTEGER, amount=Coins.__schema__)

    gas: int = 0
    amount: Coins = field(default_factory=Coins)

    def __post_init__(self):
        self.amount = Coins(self.amount)

    def to_data(self) -> dict:
        return {"gas": str(self.gas), "amount": self.amount}

    @property
    def min_gas_prices(self) -> Coins:
        return self.amount.dec_coins / self.gas

    @classmethod
    def make(cls, gas: int = 0, **denoms):
        amount = Coins()
        for denom in denoms:
            amount += Coin(denom, denoms[denom])
        return cls(gas=gas, amount=amount)

    @classmethod
    def from_data(cls, data: dict) -> StdFee:
        return cls(gas=int(data["gas"]), amount=Coins.from_data(data["amount"]))


@dataclass
class StdSignature(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(signature=S.STRING, pub_key=PublicKey.__schema__)

    signature: str
    pub_key: PublicKey

    def __str__(self) -> str:
        return self.signature

    @classmethod
    def from_data(cls, data: dict) -> StdSignature:
        if data is None:
            return None
        return cls(
            signature=data.get("signature"),
            pub_key=PublicKey.from_data(data.get("pub_key")),
        )


@dataclass
class StdTx(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^core/StdTx\Z"),
        value=S.OBJECT(
            fee=StdFee.__schema__,
            msg=S.ARRAY(S.ANY(*(mt.__schema__ for mt in MSG_TYPES.values()))),
            signatures=S.ARRAY(StdSignature.__schema__),
            memo=S.STRING,
        ),
    )

    # NOTE: msg is not plural, and is NOT a typo. This may change later for consistency.
    fee: Optional[StdFee] = None
    msg: List[StdMsg] = field(default_factory=list)
    signatures: List[StdSignature] = field(default_factory=list)
    memo: str = ""

    def to_data(self) -> dict:
        return {
            "type": "core/StdTx",
            "value": dict(self.__dict__),
        }

    @classmethod
    def from_data(cls, data: dict) -> StdTx:
        data = data["value"]
        fee = StdFee.from_data(data["fee"])
        # deserialize the messages
        msg = []
        for m in data["msg"]:
            msg_type = MSG_TYPES[m["type"]]
            msg.append(msg_type.from_data(m))
        signatures = [StdSignature.from_data(s) for s in data["signatures"]]
        return cls(fee=fee, msg=msg, signatures=signatures, memo=data["memo"])


@dataclass
class StdSignMsg(JsonSerializable):

    # TODO: Add deserialization?

    chain_id: Optional[str] = None
    account_number: Optional[int] = None
    sequence: Optional[int] = None
    fee: Optional[StdFee] = None
    msgs: List[StdMsg] = field(default_factory=list)
    memo: str = ""

    def to_tx(self) -> StdTx:
        """Get the associated `StdTx` value of the sign message, with the
        `signatures` attribute set to `None`.
        """
        return StdTx(fee=self.fee, msg=self.msgs, signatures=None, memo=self.memo)

    def to_data(self) -> dict:
        d = dict(self.__dict__)
        d["account_number"] = str(self.account_number)
        d["sequence"] = str(self.sequence)
        return d


# Deprecated -- see MsgInfo
# @dataclass
# class TxLogEntry(JsonSerializable):
# __schema__ = S.OBJECT(
#     msg_index=S.INTEGER,
#     success=S.BOOLEAN,
#     log=S.STRING,
#     events=S.ARRAY(
#         S.OBJECT(
#             type=S.STRING,
#             attributes=S.ARRAY(S.OBJECT(key=S.STRING, value=S.STRING)),
#         )
#     ),
# )

# msg: StdMsg
# success: bool
# log: str
# events: dict

# def __post_init__(self):
#     try:
#         log = json.loads(self.log)
#         if type(log) == dict:
#             self.log = JiguBox(log)
#     except:
#         log = None


@dataclass
class TxInfo(JsonSerializable, JsonDeserializable):
    """Holds data about a transaction that has been broadcasted and included in a block."""

    # TODO: Add Schema
    __schema__ = {}

    height: int
    txhash: str
    # logs: List[TxLogEntry] -- deprecated, log information uses MsgInfo right in msg now!
    gas_wanted: int
    gas_used: int
    timestamp: Timestamp

    # Merged with StdTx...
    # The rearrangement here from StdTx's arg list structure is
    # for better pretty-printing, and the fact that nobody will
    # be making TxInfo objects manually.
    fee: StdFee
    memo: str
    signatures: List[StdSignature]
    msg: List[MsgInfo]

    @property
    def tx(self):
        return StdTx(
            fee=self.fee, msg=self.msg, signatures=self.signatures, memo=self.memo
        )

    @property
    def msgs(self):
        # TODO: make the querying more powerful and extend to events
        """This is an alias because StdTx uses .msg whereas it makes more sense to use msgs."""
        return MsgInfosQuery(self.msg)

    @property
    def pretty_data(self):
        d = dict(self.__dict__)
        d.pop("msg")
        items = list(d.items())
        items.append(("msgs", self.msgs))
        return items

    def to_data(self) -> dict:
        logs = []
        for i, msginfo in enumerate(self.msg):  # treat as a log
            logs.append(
                {
                    "msg_index": i,
                    "success": msginfo.success,
                    "log": json.dumps(msginfo.log),
                    "events": [event for l in msginfo.events.values() for event in l],
                }
            )
            print(logs)

        print(logs)
        return {
            "height": self.height,
            "txhash": self.txhash,
            "logs": logs,
            "gas_wanted": self.gas_wanted,
            "gas_used": self.gas_used,
            "timestamp": self.timestamp,
            "tx": self.tx,
        }

    @classmethod
    def from_data(cls, data: dict) -> TxInfo:
        tx = data.get("tx") and StdTx.from_data(data["tx"])
        # need to merge tx.msg and logs to get something useful!
        # schema(log) = {msg_index: int, success: bool, log: str, events: event[]}
        # schema(event) = {attributes: [{key: ..., value: ...}], type: str}
        logs = TxInfo.merge_logs(data, tx)
        return cls(
            height=data.get("height") and int(data["height"]),
            txhash=data.get("txhash"),
            # logs=None,
            gas_wanted=data.get("gas_wanted") and int(data["gas_wanted"]),
            gas_used=data.get("gas_used") and int(data["gas_used"]),
            timestamp=data.get("timestamp") and Timestamp.from_data(data["timestamp"]),
            fee=tx.fee,
            msg=logs,
            signatures=tx.signatures,
            memo=tx.memo,
        )

    @staticmethod
    def merge_logs(data: dict, tx: StdTx) -> List[MsgInfo]:
        """Joins logs and tx data together so that you can access log data straight
        from the MsgInfo."""
        temp_logs = data.get("logs")
        if temp_logs is None:
            logs = None
        else:
            logs = list()
            for i, l in enumerate(temp_logs):
                events = [Event.from_data(e) for e in l["events"]]
                logs.append(
                    MsgInfo(
                        msg=tx.msg[i],
                        success=l["success"],
                        log=l["log"],
                        events=EventsQuery(events),
                    ),
                )
        return logs


@dataclass
class TxBroadcastResult(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        height=S.STRING_INTEGER,
        txhash=S.STRING,
        raw_log=S.STRING,
        logs=S.ARRAY(MsgInfo.__schema__),
        gas_wanted=S.STRING_INTEGER,
        gas_used=S.STRING_INTEGER,
        events=S.ARRAY({}),
    )

    height: int
    txhash: str
    raw_log: str
    # logs: list
    gas_wanted: int
    gas_used: int
    # events: list
    msgs: MsgInfosQuery

    @property
    def pretty_data(self):
        d = dict(self.__dict__)
        d.pop("raw_log")
        return d.items()

    @property
    def events(self):
        return self.msgs.events

    @classmethod
    def from_data(cls, data: dict, tx) -> TxBroadcastResult:
        logs = TxInfo.merge_logs(data, tx)
        if logs:
            logs = MsgInfosQuery(logs)
        return cls(
            height=data.get("height") and int(data["height"]),
            txhash=data.get("txhash"),
            raw_log=data.get("raw_log"),
            gas_wanted=data.get("gas_wanted") and int(data["gas_wanted"]),
            gas_used=data.get("gas_used") and int(data["gas_used"]),
            # events=data.get("events"),
            msgs=logs,
        )



from typing import List

from jigu.core import Coin, ExchangeRatePrevote, ExchangeRateVote


class DenomQuery(object):
    def __init__(self, terra, denom: str):
        self.terra = terra
        self.denom = denom

    def __repr__(self):
        return f"DenomQuery({self.denom!r}) -> {self.terra}"

    def prevotes(self) -> List[ExchangeRatePrevote]:
        return self.terra.oracle.prevotes(denom=self.denom)

    def votes(self) -> List[ExchangeRateVote]:
        return self.terra.oracle.votes(denom=self.denom)

    def exchange_rate(self) -> Coin:
        return self.terra.oracle.exchange_rate(denom=self.denom)

    def supply(self) -> Coin:
        return self.terra.supply(denom=self.denom)

    def community_pool(self) -> Coin:
        return self.terra.distribution.community_pool(denom=self.denom)



import os

import bech32

from jigu.error import (
    InvalidAccAddress,
    InvalidAccPubKey,
    InvalidValAddress,
    InvalidValConsAddress,
    InvalidValConsPubKey,
    InvalidValPubKey,
    ValidationError,
)


def is_bech32(data: str) -> bool:
    # ignore if we are testing
    if os.environ.get("JIGU_BECH32_VALIDATE", "1") == "0":
        return True
    return bech32.bech32_decode(data) != (None, None,)


def is_acc_address(data: str) -> bool:
    return (
        is_bech32(data)
        and data.startswith("terra")
        and not data.startswith("terravaloper")
        and not data.startswith("terravalcons")
        and not data.startswith("terrapub")
        and not data.startswith("terravalpub")
        and not data.startswith("terravalconspub")
    )


def is_val_address(data: str) -> bool:
    return (
        is_bech32(data)
        and data.startswith("terravaloper")
        and not data.startswith("terravaloperpub")
    )


def is_val_consaddress(data: str) -> bool:
    return (
        is_bech32(data)
        and data.startswith("terravalcons")
        and not data.startswith("terravalconspub")
    )


def is_acc_pubkey(data: str) -> bool:
    return is_bech32(data) and data.startswith("terrapub")


def is_val_pubkey(data: str) -> bool:
    return is_bech32(data) and data.startswith("terravalpub")


def is_val_conspubkey(data: str) -> bool:
    return is_bech32(data) and data.startswith("terravalconspub")


def validate_acc_address(data) -> str:
    """Tries to exctract an account address from the object."""
    if hasattr(data, "acc_address"):
        data = data.acc_address
    elif hasattr(data, "address"):
        data = data.address
    if not is_acc_address(data):
        raise InvalidAccAddress(data)
    return data


def validate_val_address(data) -> str:
    """Tries to exctract a validator operator address from the object."""
    if hasattr(data, "val_address"):
        data = data.val_address
    if not is_val_address(data):
        raise InvalidValAddress(data)
    return data


def validate_val_consaddress(data: str) -> str:
    if not is_val_consaddress(data):
        raise InvalidValConsAddress(data)
    return data


def validate_acc_pubkey(data: str) -> str:
    if not is_acc_pubkey(data):
        raise InvalidAccPubKey(data)
    return data


def validate_val_pubkey(data: str) -> str:
    if not is_val_pubkey(data):
        raise InvalidValPubKey(data)
    return data


def validate_val_conspubkey(data: str) -> str:
    if not is_val_conspubkey(data):
        raise InvalidValConsPubKey(data)
    return data


def validate_same_denom(d1: str, d2: str) -> None:
    if not d1 == d2:
        raise ValidationError(f"Denom '{d1}' != '{d2}', aborting.")


class Schemas(object):
    ACC_ADDRESS = {"type": "string", "pattern": r"^terra[a-zA-Z0-9]{39}\Z"}
    VAL_ADDRESS = {"type": "string", "pattern": r"^terravaloper[a-zA-Z0-9]{39}\Z"}
    VAL_CONSADDRESS = {"type": "string", "pattern": r"^terravalcons[a-zA-Z0-9]{39}\Z"}
    ACC_PUBKEY = {"type": "string", "pattern": r"^terrapub[a-zA-Z0-9]{67}\Z"}
    VAL_PUBKEY = {"type": "string", "pattern": r"^terravalpub[a-zA-Z0-9]{67}\Z"}
    VAL_CONSPUBKEY = {"type": "string", "pattern": r"^terravalconspub[a-zA-Z0-9]{67}\Z"}

    STRING = {"type": "string"}
    STRING_INTEGER = {"type": "string", "pattern": r"^[\+\-]?[0-9]+\Z"}
    INTEGER = {"type": "integer"}
    BOOLEAN = {"type": "boolean"}
    OBJ = {"type": "object"}  # generic object

    @staticmethod
    def STRING_WITH_PATTERN(pattern: str):
        return {"type": "string", "pattern": pattern}

    @staticmethod
    def STRING_ENUM(*options):
        return {"type": "string", "enum": options}

    @staticmethod
    def ANY(*schemas):
        return {"anyOf": list(schemas)}

    @staticmethod
    def ONE(*schemas):
        return {"oneOf": list(schemas)}

    @staticmethod
    def ALL(*schemas):
        return {"allOf": list(schemas)}

    @staticmethod
    def OPTIONAL(schema):
        return {"anyOf": [schema, {"type": "null"}]}

    @staticmethod
    def OBJECT(**kwargs) -> object:
        return {
            "type": "object",
            "properties": kwargs,
            "required": list(kwargs.keys()),
        }

    @staticmethod
    def ARRAY(schema, **kwargs):
        return {"type": "array", "items": schema, **kwargs}



# Request middlewares
import json
import re

import requests

from jigu.error import BadRequest, LcdInternalError, get_codespace_error, get_rpc_error
from jigu.util.serdes import serialize_to_json

from .lcdrequest import LcdRequest

__all__ = [
    "set_timeout",
    "set_query",
    "handle_codespace_errors",
    "handle_general_errors",
]


def set_timeout(client, request: requests.Request):
    kwargs = request.kwargs
    kwargs["timeout"] = kwargs.get("timeout", client.timeout)
    return LcdRequest(request.method, request.url, kwargs)


def set_query(client, request: requests.Request):
    """Add `params` to GET requests to make queries, and `data` to POST requests for
    submitting JSON data. Applies serialization strategy to JSON post requests."""

    method = request.method
    kwargs = request.kwargs
    kwargs["data"] = kwargs.get("data", None)
    kwargs["headers"] = kwargs.get("headers", {})

    if kwargs["data"] and method == "get":
        kwargs["params"] = kwargs["data"]
        del kwargs["data"]

    if method == "post":
        kwargs["headers"]["content-type"] = "application/json"
        kwargs["data"] = serialize_to_json(kwargs["data"])  # apply custom serialization
    return LcdRequest(method, request.url, kwargs)


# Response middlewares
def handle_codespace_errors(client, response: requests.Response):
    """Try to extract more specific codespace errors."""
    error_type = str(response.status_code)[0]
    if error_type in ["4", "5"]:
        try:
            res = response.json()
            error = res.get("error", "")
            if isinstance(error, dict):
                error_msg = error["message"]
            else:
                error_msg = error
        except (ValueError, TypeError, KeyError):
            error_msg = response.content
        try:
            e = json.loads(error_msg)
            if isinstance(e, dict) and "codespace" in e:
                raise get_codespace_error(e["codespace"], e["code"], e["message"])
        except json.decoder.JSONDecodeError:
            pass  # continue normally
    return response


def handle_general_errors(client, response: requests.Response):
    """Handles 4xx and 5xx errors. Checks first if they originally from an RPC error."""
    error_type = str(response.status_code)[0]
    if error_type in ["4", "5"]:
        try:
            res = response.json()
            error = res.get("error", {})
            if isinstance(error, dict):
                error_msg = error["message"]
            else:
                error_msg = error
        except (ValueError, TypeError, KeyError):
            error_msg = response.content
        match = re.match(r"^.*RPC error (\-?\d+) - (.*)$", str(error_msg))
        if match:
            raise get_rpc_error(int(match.group(1)), match.group(2), response.request)
        if error_type == "4":
            raise BadRequest(response.status_code, error_msg, response.request)
        if error_type == "5":
            raise LcdInternalError(response.status_code, error_msg, response.request)
    return response



from decimal import Decimal

import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.msg.oracle import (
    MsgDelegateFeedConsent,
    MsgExchangeRatePrevote,
    MsgExchangeRateVote,
)
from jigu.core.sdk import Coin, Dec
from jigu.error import InvalidAccAddress, InvalidValAddress, ValidationError
from jigu.util.validation import is_acc_address, is_val_address
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the __schema__ validation set for oracle messages."""
    return load_msg_examples(
        tdd,
        msg_types=[MsgDelegateFeedConsent, MsgExchangeRateVote, MsgExchangeRatePrevote],
    )


class TestMsgExchangeRateVote:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgExchangeRateVote]:
            assert_serdes_consistent(MsgExchangeRateVote, m)
            assert_serdes_exact(MsgExchangeRateVote, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgExchangeRateVote.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgExchangeRateVote, m)

    def test_matches_meta(self):
        assert MsgExchangeRateVote.type == "oracle/MsgExchangeRateVote"
        assert MsgExchangeRateVote.action == "exchangeratevote"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, val_address, other):
        """MsgExchangeRateVote should validate `feeder` to be AccAddress and `validator`
        to be ValAddress.
        """
        assume(not is_val_address(other) and not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgExchangeRateVote(
                exchange_rate=Coin("ukrw", 1000),
                salt="",
                denom="ukrw",
                feeder=other,
                validator=val_address,
            )
        with pytest.raises(InvalidValAddress):
            MsgExchangeRateVote(
                exchange_rate=Coin("ukrw", 1000),
                salt="",
                denom="ukrw",
                feeder=acc_address,
                validator=other,
            )
        A = MsgExchangeRateVote(
            exchange_rate=Coin("ukrw", 1000),
            salt="",
            denom="ukrw",
            feeder=acc_address,
            validator=val_address,
        )
        assert A.feeder == acc_address
        assert A.validator == val_address

    @pytest.mark.parametrize("t", [int, Dec, Decimal, str, lambda x: Coin("ukrw", x)])
    def test_constructor_canonizes_exchange_rate_to_coin(
        self, acc_address, val_address, t
    ):
        A = MsgExchangeRateVote(
            exchange_rate=t(1000),
            salt="",
            denom="ukrw",
            feeder=acc_address,
            validator=val_address,
        )

        assert isinstance(A.exchange_rate, Coin)
        assert A.exchange_rate.denom == "ukrw"

    def test_exchange_rate_wrong_denom(self, acc_address, val_address):
        """Passing in `exchange_rate` as `Coin` will require both the `Coin.denom` and `denom`
        parameter to match.
        """

        with pytest.raises(ValidationError):
            MsgExchangeRateVote(
                exchange_rate=Coin("ukrw", 1000),
                salt="",
                denom="uusd",
                feeder=acc_address,
                validator=val_address,
            )

        A = MsgExchangeRateVote(
            exchange_rate=Coin("ukrw", 1000),
            salt="",
            denom="ukrw",
            feeder=acc_address,
            validator=val_address,
        )

        assert A.exchange_rate.denom == A.denom

    def test_vote_hash(self):

        msg1 = MsgExchangeRateVote(
            exchange_rate=Coin("umnt", "603.899000000000000000"),
            salt="0dff",
            denom="umnt",
            feeder="terra13ld7qfuq37328mw6f5kunez3e2ygqumxfcysms",
            validator="terravaloper1vqnhgc6d0jyggtytzqrnsc40r4zez6tx99382w",
        )

        assert msg1.vote_hash == "b338c8a65a132edcf3e9a13013997cdf1e2b283d"


class TestMsgExchangeRatePrevote:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgExchangeRatePrevote]:
            assert_serdes_consistent(MsgExchangeRatePrevote, m)
            assert_serdes_exact(MsgExchangeRatePrevote, m)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgExchangeRatePrevote.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgExchangeRatePrevote, m)

    def test_matches_meta(self):
        assert MsgExchangeRatePrevote.type == "oracle/MsgExchangeRatePrevote"
        assert MsgExchangeRatePrevote.action == "exchangerateprevote"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, val_address, other):
        """`MsgExchangeRateVote` should validate `feeder` to be AccAddress and `validator`
        to be ValAddress.
        """
        assume(not is_val_address(other) and not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgExchangeRatePrevote(
                hash="", denom="", feeder=other, validator=val_address,
            )
        with pytest.raises(InvalidValAddress):
            MsgExchangeRatePrevote(
                hash="", denom="", feeder=acc_address, validator=other,
            )
        A = MsgExchangeRatePrevote(
            hash="", denom="", feeder=acc_address, validator=val_address,
        )
        assert A.feeder == acc_address
        assert A.validator == val_address


class TestMsgDelegateFeedConsent:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgDelegateFeedConsent]:
            assert_serdes_consistent(MsgDelegateFeedConsent, m)
            assert_serdes_exact(MsgDelegateFeedConsent, m)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgDelegateFeedConsent.__schema__))
    def test_schema(self, m):
        assert_serdes_consistent(MsgDelegateFeedConsent, m)

    def test_matches_meta(self):
        assert MsgDelegateFeedConsent.type == "oracle/MsgDelegateFeedConsent"
        assert MsgDelegateFeedConsent.action == "delegatefeeder"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, val_address, other):
        """MsgDelegateFeedConsent should validate `operator` to be ValAddress and `delegate`
        to be AccAddress.
        """
        assume(not is_val_address(other) and not is_acc_address(other))
        with pytest.raises(InvalidValAddress):
            MsgDelegateFeedConsent(operator=other, delegate=acc_address)
        with pytest.raises(InvalidAccAddress):
            MsgDelegateFeedConsent(operator=val_address, delegate=other)
        A = MsgDelegateFeedConsent(operator=val_address, delegate=acc_address)
        assert A.operator == val_address
        assert A.delegate == acc_address



from typing import Optional, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import Coin, Coins, Dec, PolicyConstraints
from jigu.core.denoms import uLuna
from jigu.util.serdes import JiguBox

__all__ = ["TreasuryApi"]


class TreasuryApi(BaseApi):
    def tax_cap(self, denom: str) -> Union[ApiResponse, Coin]:
        res = self._api_get(f"/treasury/tax_cap/{denom}")
        return project(res, Coin(denom, res))

    def tax_rate(self) -> Union[ApiResponse, Dec]:
        res = self._api_get("/treasury/tax_rate")  # tr
        return project(res, Dec.deserialize(res))

    def reward_weight(self) -> Union[ApiResponse, Dec]:
        res = self._api_get("/treasury/reward_weight")  # rw
        return project(res, Dec.deserialize(res))

    def tax_proceeds(self, denom: Optional[str] = None) -> Union[ApiResponse, Coins]:
        res = self._api_get("/treasury/tax_proceeds")
        tax_proceeds = Coins.deserialize(res)
        return project(res, tax_proceeds[denom] if denom else tax_proceeds)

    def seigniorage_proceeds(self) -> Union[ApiResponse, Coin]:
        res = self._api_get("/treasury/seigniorage_proceeds")
        return project(res, Coin(uLuna, int(res)))

    def params(self, key: Optional[str] = None) -> Union[ApiResponse, Dec, int, dict]:
        res = self._api_get("/treasury/parameters")
        p = JiguBox(
            res,
            box_recast={
                "seigniorage_burden_target": Dec,
                "mining_increment": Dec,
                "window_short": int,
                "window_long": int,
                "window_probation": int,
            },
        )
        p["tax_policy"] = PolicyConstraints.deserialize(p["tax_policy"])
        p["reward_policy"] = PolicyConstraints.deserialize(p["reward_policy"])
        return project(res, p[key] if key else p)



from __future__ import annotations

import json
import re
from dataclasses import dataclass, field

from bidict import bidict

from jigu.core.proposal import Content
from jigu.core.sdk import Coins, Dec
from jigu.core.treasury import PolicyConstraints
from jigu.error import InvalidParamChange
from jigu.util.serdes import (
    JiguBox,
    JsonDeserializable,
    JsonSerializable,
    serialize_to_json,
)
from jigu.util.validation import Schemas as S

__all__ = ["ParamChanges", "ParameterChangeProposal"]

symbol = re.compile(r"^[a-zA-Z_][a-zA-Z_0-9]*$")

ParamChangeSchema = S.OBJECT(subspace=S.STRING, key=S.STRING, value=S.STRING)

# For each subspace, map JSON-param key to (ParamStore key, deserializing-fn, serialiazing-fn)
# Serializing function is necessary due to weird ways Cosmos's params module treats integers.

# TODO: This file could use some work.

# Define the deserialization and serialization functions for governance params
def deserialize_deposit(data: dict):
    if data.get("min_deposit"):
        data["min_deposit"] = Coins.deserialize(data["min_deposit"])
    if data.get("max_deposit_period"):
        data["max_deposit_period"] = int(data["max_deposit_period"])
    return data


def serialize_deposit(data: dict):
    if data.get("max_deposit_period"):
        data["max_deposit_period"] = str(int(data["max_deposit_period"]))
    return data


def deserialize_voting(data: dict):
    if data.get("voting_period"):
        data["voting_period"] = int(data["voting_period"])
    return data


def serialize_voting(data: dict):
    if data.get("voting_period"):
        data["voting_period"] = str(int(data["voting_period"]))
    return data


def deserialize_tally(data: dict):
    if data.get("quorum"):
        data["quorum"] = Dec(data["quorum"])
    if data.get("threshold"):
        data["threshold"] = Dec(data["threshold"])
    if data.get("veto"):
        data["veto"] = Dec(data["veto"])
    return data


def serialize_tally(data: dict):
    return data


# TODO: this could be refactored into XXXModuleParams class in core, with keys
# as attributes, and API requests for Module Params give you an instance of
# ModuleParams with all keys loaded with their current values. Then, to build
# param changes, you would instantiate an empty ModuleParams class with the
# values initialized to ModuleParams.EMPTY and then set each one.

PARAM_DEFNS = {
    "distribution": {
        "community_tax": ("communitytax", Dec),
        "base_proposer_reward": ("baseproposerreward", Dec),
        "bonus_proposer_reward": ("bonusproposerreward", Dec),
        "withdraw_addr_enabled": ("withdrawaddrenabled", bool),
    },
    "staking": {
        "unbonding_time": ("UnbondingTime", int, str),
        "max_validators": ("MaxValidators", int, int),
        "max_entries": ("KeyMaxEntries", int, int),
        "bond_denom": ("BondDenom", str),
    },
    "slashing": {
        "max_evidence_age": ("MaxEvidenceAge", int, str),  # no longer in cosmos master/
        "signed_blocks_window": ("SignedBlocksWindow", int, str),
        "min_signed_per_window": ("MinSignedPerWindow", Dec),
        "downtime_jail_duration": ("DowntimeJailDuration", int, str),
        "slash_fraction_double_sign": ("SlashFractionDoubleSign", Dec),
        "slash_fraction_downtime": ("SlashFractionDowntime", Dec),
    },
    "oracle": {
        "vote_period": ("voteperiod", int, str),
        "vote_threshold": ("votethreshold", Dec),
        "reward_band": ("rewardband", Dec),
        "reward_distribution_window": ("rewarddistributionwindow", int, str),
        "whitelist": ("whitelist", None),
        "slash_fraction": ("slashfraction", Dec),
        "slash_window": ("slashwindow", int, str),
        "min_valid_per_window": ("minvalidperwindow", Dec),
    },
    "market": {
        "pool_recovery_period": ("poolrecoveryperiod", int, str),
        "base_pool": ("basepool", Dec),
        "min_spread": ("minspread", Dec),
        "illiquid_tobin_tax_list": ("illiquidtobintaxlist", None),
    },
    "treasury": {
        "tax_policy": ("taxpolicy", PolicyConstraints.deserialize),
        "reward_policy": ("rewardpolicy", PolicyConstraints.deserialize),
        "min_spread": ("minspread", Dec),
        "seigniorage_burden_target": ("seigniorageburdentarget", Dec),
        "mining_increment": ("miningincrement", Dec),
        "window_short": ("windowshort", int, str),
        "window_long": ("windowlong", int, str),
        "window_probation": ("windowprobation", int, str),
    },
    "gov": {
        "deposit_params": ("depositparams", deserialize_deposit, serialize_deposit),
        "voting_params": ("votingparams", deserialize_voting, serialize_voting),
        "tally_params": ("tallyparams", deserialize_tally, serialize_tally),
    },
}

# create lookup table for deserialization
DES_LOOKUP_TABLE = {}
for subspace, keys in PARAM_DEFNS.items():
    DES_LOOKUP_TABLE[subspace] = {d[0]: d[1] for k, d in keys.items()}

# DES_LOOKUP_TABLE[subspace][paramkey aka d[0]] = JSON param key
# DES_LOOKUP_TABLE["treasury"]["windowlong"] -> "window_long"

# create lookup table for serialization
PARAMSTORE_KEY_LOOKUP_TABLE = {}
for subspace, keys in PARAM_DEFNS.items():
    PARAMSTORE_KEY_LOOKUP_TABLE[subspace] = bidict({k: d[0] for k, d in keys.items()})


class ParamChanges(JsonSerializable, JsonDeserializable):

    __schema__ = S.ARRAY(ParamChangeSchema)

    def __init__(self, changes: dict):
        for k, v in changes.items():
            m = symbol.match(k)
            if not m:
                raise InvalidParamChange(
                    f"Parameter change subspace could not be parsed: {k}"
                )
            if not isinstance(v, dict):
                raise InvalidParamChange(
                    f"Parameter change value should be a dict but got: '{type(v)}' for {k}"
                )
            for sk, sv in v.items():
                sm = symbol.match(sk)
                if not sm:
                    raise InvalidParamChange(
                        f"Parameter change key could not be parsed - {k}: '{sk}'"
                    )
        self.changes = JiguBox(changes)

    def __repr__(self) -> str:
        return f"<ParamChanges {self.changes!r}>"

    def __getattr__(self, name: str) -> JiguBox:
        return self.changes[name]

    def __getitem__(self, item) -> JiguBox:
        return self.changes[item]

    @staticmethod
    def _get_key(subspace, key, inverse=False):
        """Each parameter has a special key in the ParamStore that might not correspond
        to the parameter's JSON-name. We use this function to look up the ParamStore key
        for the JSON-name, which we are familiar with.

        Set `inverse` to `True` to look up from the opposite direction for deserialization.
        """
        try:
            table = PARAMSTORE_KEY_LOOKUP_TABLE[subspace]
            if inverse:
                table = table.inverse
            return table[key]
        except KeyError:
            return key

    @staticmethod
    def _marshal_value(subspace, key, value):
        """Formats the value for param change. Terra node expects all the values to be
        JSON-encoded, and Amino:JSON int/int64/uint/uint64 expects quoted values for
        JavaScript numeric support, and int16/uint16 need `int`.
        """
        try:
            if key not in PARAM_DEFNS[subspace]:  # if not JSON-name
                # try paramstore name
                key = ParamChanges._get_key(subspace, key, inverse=True)
            if len(PARAM_DEFNS[subspace][key]) == 3:
                value = PARAM_DEFNS[subspace][key][2](value)
        except KeyError:
            pass
        return serialize_to_json(value)

    @staticmethod
    def _unmarshal_value(subspace, key, value):
        """Looks up the correct type to decode the right type for the parameter change."""
        if subspace in DES_LOOKUP_TABLE and key in DES_LOOKUP_TABLE[subspace]:
            if DES_LOOKUP_TABLE[subspace][key] is not None:
                return DES_LOOKUP_TABLE[subspace][key](value)
        return value

    @property
    def pretty_data(self):
        return self.changes.items()

    def to_data(self) -> list:
        param_changes = []
        for subspace, v in self.changes.items():
            for key, value in v.items():
                param_changes.append(
                    {
                        "subspace": subspace,
                        "key": self._get_key(subspace, key),
                        "value": self._marshal_value(subspace, key, value),
                    }
                )
        return param_changes

    @classmethod
    def from_data(cls, data: list) -> ParamChanges:
        changes = JiguBox(default_box=True)
        for p in data:
            subspace = p["subspace"]
            key = cls._get_key(
                subspace, p["key"], inverse=True
            )  # p["key"] is paramstore key, we are using json-name keys inside Jigu
            value = cls._unmarshal_value(subspace, p["key"], json.loads(p["value"]))
            changes[subspace][key] = value
        return cls(changes)


@dataclass
class ParameterChangeProposal(Content):

    type = "params/ParameterChangeProposal"
    ParamChanges = ParamChanges  # alias for easy access

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^params/ParameterChangeProposal\Z"),
        value=S.OBJECT(
            title=S.STRING, description=S.STRING, changes=ParamChanges.__schema__,
        ),
    )

    title: str
    description: str
    changes: ParamChanges = field(default_factory=ParamChanges)

    def __post_init__(self):
        # validation checks for key
        if isinstance(self.changes, dict):
            self.changes = ParamChanges(self.changes)

    @classmethod
    def from_data(cls, data: dict) -> ParameterChangeProposal:
        data = data["value"]
        return cls(
            title=data["title"],
            description=data["description"],
            changes=ParamChanges.from_data(data["changes"]),
        )



MicroLunaDenom = "uluna"
MicroUSDDenom = "uusd"
MicroKRWDenom = "ukrw"
MicroSDRDenom = "usdr"
MicroCNYDenom = "ucny"
MicroJPYDenom = "ujpy"
MicroEURDenom = "ueur"
MicroGBPDenom = "ugbp"
MicroMNTDenom = "umnt"

uLuna = MicroLunaDenom
uUSD = MicroUSDDenom
uKRW = MicroKRWDenom
uSDR = MicroSDRDenom
uCNY = MicroCNYDenom
uJPY = MicroJPYDenom
uEUR = MicroEURDenom
uGBP = MicroGBPDenom
uMNT = MicroMNTDenom

STAKING = uLuna
STANDARD = uSDR



"""Tools for pretty-printing."""
import textwrap
from typing import Any, List, Tuple

import tabulate

import jigu
import jigu.util.serdes

__all__ = [
    "MAX_ROWHEADER_LEN",
    "MAX_ROWSTRING_LEN",
    "see",
    "pretty_repr",
    "PrettyPrintable",
]

MAX_ROWHEADER_LEN = 28
MAX_ROWSTRING_LEN = 45

# TODO: fix display when inside ipython console
try:
    in_ipython = __IPYTHON__
except NameError:
    in_ipython = False

if in_ipython:
    style = "html"
    sep = "<br/>"
else:
    style = "fancy_grid"
    sep = "\n"


def see(obj):
    print(pretty_repr(obj))


def terminal_title(text):
    return f"[{text}]"


def html_bold(name):
    """Adds boldface if in ipython."""
    if in_ipython:
        return "<b>" + name + "</b>"
    return name


def pretty_repr(obj, path=""):
    if isinstance(obj, list):
        if all(isinstance(x, str) for x in obj):
            return sep.join(textwrap.wrap(", ".join(obj), width=MAX_ROWSTRING_LEN))
        if len(obj) >= 1:
            rst = (
                f"{sep}... +{len(obj)-1} more (see {path.rstrip('[]')})"
                if len(obj) > 1
                else sep + ""
            )
            return pretty_repr(obj[0], path=path) + rst
    if isinstance(obj, dict) and not isinstance(obj, jigu.util.serdes.JiguBox):
        return jigu.util.serdes.JiguBox(obj)._pretty_repr_(path)
    if isinstance(obj, str):
        return sep.join(textwrap.wrap(obj, width=MAX_ROWSTRING_LEN))
    if hasattr(obj, "_pretty_repr_"):
        return obj._pretty_repr_(path)
    else:
        return repr(obj)


class PrettyPrintable:
    """Mixin for pretty-printable classes, includes a default pretty-printer that
    shows a table of attributes and their values.
    """

    def _repr_html_(self) -> str:
        """Table HTML repr() for Jupyter."""
        return self._pretty_repr_()

    @property
    def pretty_data(self) -> List[Tuple[str, Any]]:
        """A list of (name, data) pairs desired for inclusion in the pretty output.
        Override this to use default pretty printer, override `_pretty_repr_` to do otherwise.
        """
        return self.__dict__.items()

    @property
    def pretty_header(self) -> str:
        """A title that will be displayed for the default pretty printer."""
        return self.__class__.__name__

    @property
    def pretty_iterable(self) -> bool:
        """Override if the object is iterable, for the default pretty printer."""
        return False

    def _pretty_output(self, path: str = "") -> List[Tuple[str, str]]:
        """Override this if you want to customize how to render child elements for
        the for the default pretty printer."""
        items = []
        for d in self.pretty_data:
            name, data = d
            revised_name = html_bold(
                name[:12] + "..." + name[-4:] if len(name) > MAX_ROWHEADER_LEN else name
            )
            if (
                isinstance(d[1], list)
                or isinstance(d[1], dict)
                or getattr(d[1], "pretty_iterable", False)
            ):
                if isinstance(d[1], list) or getattr(d[1], "pretty_iterable", False):
                    name = f"{name}[]"
                items.append((revised_name, pretty_repr(data, f"{path}.{name}")))
            else:
                items.append((revised_name, pretty_repr(data, path)))
        return items

    def _pretty_repr_(self, path: str = "") -> str:
        """Default pretty printer which formats items as a table. Override this to
        customize pretty-printing.
        """
        if path == "":  # we are at root
            path = self.__class__.__name__
        if self.pretty_header:
            title = (
                "<b><u>" + self.pretty_header + "</u></b>"
                if in_ipython
                else terminal_title(self.pretty_header)
            )
            header = title + sep
        else:
            header = ""
        return header + tabulate.tabulate(
            self._pretty_output(path), tablefmt=style, disable_numparse=True
        )

    @property
    def _pp(self):
        """Shortcut for seeing pretty-printing output."""
        see(self)
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

from jigu.core.sdk import Coins, Dec, PublicKey
from jigu.util.serdes import JiguBox, JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_acc_address

__all__ = ["Account", "VestingScheduleEntry", "LazyGradedVestingAccount"]


@dataclass
class Account(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^core/Account\Z"),
        value=S.OBJECT(
            address=S.ACC_ADDRESS,
            coins=Coins.__schema__,
            public_key=S.OPTIONAL(PublicKey.__schema__),
            account_number=S.STRING_INTEGER,
            sequence=S.STRING_INTEGER,
        ),
    )

    address: str
    coins: Coins
    public_key: Optional[PublicKey]
    account_number: int
    sequence: int

    def __post_init__(self):
        if self.address:
            self.address = validate_acc_address(self.address)
        if self.coins:
            self.coins = Coins(self.coins)
        self.account_number = int(self.account_number)
        self.sequence = int(self.sequence)

    def to_data(self) -> dict:
        d = dict(self.__dict__)
        d["account_number"] = str(self.account_number)
        d["sequence"] = str(self.sequence)
        return {"type": "core/Account", "value": d}

    @classmethod
    def from_data(cls, data: dict) -> Account:
        if "type" in data:
            data = data["value"]
        pk = data["public_key"]
        if pk is not None:
            pk = PublicKey.deserialize(data["public_key"])
        return cls(
            address=data["address"],
            coins=Coins.from_data(data["coins"]),
            public_key=pk,
            account_number=data["account_number"],
            sequence=data["sequence"],
        )


@dataclass
class VestingScheduleEntry(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        start_time=S.STRING_INTEGER, end_time=S.STRING_INTEGER, ratio=Dec.__schema__
    )

    start_time: int
    end_time: int
    ratio: Dec

    def to_data(self):
        return {
            "start_time": str(self.start_time),
            "end_time": str(self.end_time),
            "ratio": self.ratio,
        }

    @classmethod
    def from_data(cls, data: dict) -> VestingScheduleEntry:
        return cls(
            start_time=int(data["start_time"]),
            end_time=int(data["end_time"]),
            ratio=Dec.from_data(data["ratio"]),
        )


@dataclass(init=False)
class LazyGradedVestingAccount(Account):

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^core/LazyGradedVestingAccount\Z"),
        value=S.OBJECT(
            BaseVestingAccount=S.OBJECT(
                BaseAccount=Account.__schema__["properties"]["value"],
                original_vesting=Coins.__schema__,
                delegated_free=Coins.__schema__,
                delegated_vesting=Coins.__schema__,
                end_time=S.STRING_INTEGER,
            ),
            vesting_schedules=S.ARRAY(
                S.OBJECT(
                    denom=S.STRING, schedules=S.ARRAY(VestingScheduleEntry.__schema__)
                )
            ),
        ),
    )

    original_vesting: Coins
    delegated_free: Coins
    delegated_vesting: Coins
    end_time: int
    vesting_schedules: Dict[str, List[VestingScheduleEntry]]

    def __init__(
        self,
        base_account: Account,
        original_vesting: Coins,
        delegated_free: Coins,
        delegated_vesting: Coins,
        end_time: int,
        vesting_schedules: Dict[str, List[VestingScheduleEntry]],
    ):
        Account.__init__(
            self,
            address=base_account.address,
            coins=base_account.coins,
            public_key=base_account.public_key,
            account_number=base_account.account_number,
            sequence=base_account.sequence,
        )
        self._base_account = base_account
        self.original_vesting = original_vesting
        self.delegated_free = delegated_free
        self.delegated_vesting = delegated_vesting
        self.end_time = end_time
        self.vesting_schedules = vesting_schedules

    def to_data(self):
        return {
            "type": "core/LazyGradedVestingAccount",
            "value": {
                "BaseVestingAccount": {
                    "BaseAccount": Account.to_data(self._base_account)["value"],
                    "original_vesting": self.original_vesting,
                    "delegated_free": self.delegated_free,
                    "delegated_vesting": self.delegated_vesting,
                    "end_time": str(self.end_time),
                },
                "vesting_schedules": [
                    {"denom": denom, "schedules": self.vesting_schedules[denom]}
                    for denom in self.vesting_schedules
                ],
            },
        }

    @property
    def pretty_data(self):
        d = dict(self.__dict__)
        d.pop("_base_account")
        return d.items()

    @classmethod
    def from_data(cls, data: dict) -> LazyGradedVestingAccount:
        if "type" in data:
            data = data["value"]  # disregard type
        bva = data[
            "BaseVestingAccount"
        ]  # value = { bva { ... } vesting_schedules { } }
        original_vesting = Coins.from_data(bva["original_vesting"])
        delegated_free = Coins.from_data(bva["delegated_free"])
        delegated_vesting = Coins.from_data(bva["delegated_vesting"])
        vesting_schedules = JiguBox({})
        for s in data["vesting_schedules"]:
            vesting_schedules[s["denom"]] = [
                VestingScheduleEntry.deserialize(e) for e in s["schedules"]
            ]
        return cls(
            base_account=Account.from_data(bva["BaseAccount"]),
            original_vesting=original_vesting,
            delegated_free=delegated_free,
            delegated_vesting=delegated_vesting,
            end_time=int(bva["end_time"]),
            vesting_schedules=vesting_schedules,
        )



from __future__ import annotations

from dataclasses import dataclass

from jigu.core import ValAddress
from jigu.core.msg import StdMsg
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_val_address

__all__ = ["MsgUnjail"]


@dataclass
class MsgUnjail(StdMsg):

    type = "cosmos/MsgUnjail"
    action = "unjail"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^cosmos/MsgUnjail\Z"),
        value=S.OBJECT(address=S.VAL_ADDRESS),
    )

    address: ValAddress

    def __post_init__(self):
        self.address = validate_val_address(self.address)

    @classmethod
    def from_data(cls, data: dict) -> MsgUnjail:
        data = data["value"]
        return cls(address=data["address"])



from __future__ import annotations

from dataclasses import dataclass

from jigu.core import Dec
from jigu.core.proposal import Content
from jigu.util.validation import Schemas as S

__all__ = ["TaxRateUpdateProposal"]


@dataclass
class TaxRateUpdateProposal(Content):

    type = "treasury/TaxRateUpdateProposal"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^treasury/TaxRateUpdateProposal\Z"),
        value=S.OBJECT(title=S.STRING, description=S.STRING, tax_rate=Dec.__schema__),
    )

    title: str
    description: str
    tax_rate: Dec

    @classmethod
    def from_data(cls, data: dict) -> TaxRateUpdateProposal:
        data = data["value"]
        return cls(
            title=data["title"],
            description=data["description"],
            tax_rate=Dec(data["tax_rate"]),
        )


@dataclass
class RewardWeightUpdateProposal(Content):

    type = "treasury/RewardWeightUpdateProposal"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^treasury/RewardWeightUpdateProposal\Z"),
        value=S.OBJECT(
            title=S.STRING, description=S.STRING, reward_weight=Dec.__schema__
        ),
    )

    title: str
    description: str
    reward_weight: Dec

    @classmethod
    def from_data(cls, data: dict) -> RewardWeightUpdateProposal:
        data = data["value"]
        return cls(
            title=data["title"],
            description=data["description"],
            reward_weight=Dec(data["reward_weight"]),
        )



import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema

from jigu.core.auth.account import LazyGradedVestingAccount
from testtools import assert_serdes_consistent, assert_serdes_exact, load_data


@pytest.fixture
def examples(tdd):
    return load_data(tdd, "objects/auth/LazyGradedVestingAccount.json")


class TestVestingAccountSerdes:
    def test_schema_valid(self, examples):
        for acc in examples:
            assert_serdes_consistent(LazyGradedVestingAccount, acc)
            assert_serdes_exact(LazyGradedVestingAccount, acc)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(acc=from_schema(LazyGradedVestingAccount.__schema__))
    def test_schema(self, acc):
        assert_serdes_consistent(LazyGradedVestingAccount, acc)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

from jigu.core import AccAddress, Coin, Coins
from jigu.core.bank import Input, Output
from jigu.core.msg import StdMsg
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_acc_address

__all__ = ["MsgSend", "MsgMultiSend"]


@dataclass
class MsgSend(StdMsg):

    type = "bank/MsgSend"
    action = "send"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^bank/MsgSend\Z"),
        value=S.OBJECT(
            from_address=S.ACC_ADDRESS,
            to_address=S.ACC_ADDRESS,
            amount=Coins.__schema__,
        ),
    )

    from_address: AccAddress
    to_address: AccAddress
    amount: Union[Coins, List[Coin]]

    def __post_init__(self):
        self.from_address = validate_acc_address(self.from_address)
        self.to_address = validate_acc_address(self.to_address)
        self.amount = Coins(self.amount)

    @classmethod
    def from_data(cls, data: dict) -> MsgSend:
        data = data["value"]
        amount = Coins.from_data(data["amount"])
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            amount=amount,
        )


@dataclass(init=False)
class MsgMultiSend(StdMsg):

    type = "bank/MsgMultiSend"
    action = "multisend"

    Input = Input  # alias
    Output = Output  # alias

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^bank/MsgMultiSend\Z"),
        value=S.OBJECT(
            inputs=S.ARRAY(Input.__schema__), outputs=S.ARRAY(Output.__schema__)
        ),
    )

    inputs: List[Union[Input, dict]]
    outputs: List[Union[Output, dict]]

    def __init__(
        self, inputs: List[Input], outputs: List[Output],
    ):
        self.inputs = []
        self.outputs = []

        for i in inputs:
            if isinstance(i, dict):
                self.inputs.append(Input(address=i["address"], coins=i["coins"]))
            elif isinstance(i, Input):
                self.inputs.append(i)
            else:
                raise TypeError(
                    f"invalid item {type(i)} encountered in MsgMultiSend.inputs, can only accept 'Input' or 'dict'."
                )
        for o in outputs:
            if isinstance(o, dict):
                self.outputs.append(Output(address=o["address"], coins=o["coins"]))
            elif isinstance(o, Output):
                self.outputs.append(o)
            else:
                raise TypeError(
                    f"invalid item {type(o)} encountered in MsgMultiSend.outputs, can only accept 'Output' or 'dict'."
                )

    @classmethod
    def from_data(cls, data: dict) -> MsgMultiSend:
        data = data["value"]
        return cls(
            inputs=[Input.from_data(i) for i in data["inputs"]],
            outputs=[Output.from_data(o) for o in data["outputs"]],
        )



import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.bank import Input, Output
from jigu.core.msg.bank import MsgMultiSend, MsgSend
from jigu.core.sdk import Coin, Coins
from jigu.error import InvalidAccAddress
from jigu.util.validation import is_acc_address
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the __schema__ validation set for slashing messages."""
    return load_msg_examples(tdd, msg_types=[MsgSend, MsgMultiSend],)


@pytest.fixture
def dict_list(acc_address):
    return [
        {"address": acc_address, "coins": []},
        {"address": acc_address, "coins": []},
    ]


@pytest.fixture
def input_list(acc_address):
    return [
        Input(address=acc_address, coins=[]),
        Input(address=acc_address, coins=[]),
    ]


@pytest.fixture
def output_list(acc_address):
    return [
        Output(address=acc_address, coins=[]),
        Output(address=acc_address, coins=[]),
    ]


@pytest.fixture
def mixed_list(acc_address):
    return [
        Output(address=acc_address, coins=[]),
        Input(address=acc_address, coins=[]),
        {"address": acc_address, "coins": []},
    ]


@pytest.fixture
def input_list_with_dict(acc_address):
    return [
        Input(address=acc_address, coins=[]),
        {"address": acc_address, "coins": []},
    ]


@pytest.fixture
def output_list_with_dict(acc_address):
    return [
        Output(address=acc_address, coins=[]),
        {"address": acc_address, "coins": []},
    ]


class TestMsgSend:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgSend]:
            assert_serdes_consistent(MsgSend, m)
            assert_serdes_exact(MsgSend, m)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgSend.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgSend, m)

    def test_matches_meta(self):
        assert MsgSend.type == "bank/MsgSend"
        assert MsgSend.action == "send"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, other):
        """MsgSend should validate both from_address and to_address to be AccAddress.
        """

        assume(not is_acc_address(other))

        with pytest.raises(InvalidAccAddress):
            MsgSend(from_address=acc_address, to_address=other, amount=[])
        with pytest.raises(InvalidAccAddress):
            MsgSend(from_address=other, to_address=acc_address, amount=[])

        A = MsgSend(from_address=acc_address, to_address=acc_address, amount=[])
        assert A.from_address == acc_address

    def test_constructor_amount_canonizes_list_to_coins(self, acc_address):
        """Creating a MsgSend object directly through its constructor should be conveniently
        achievable with either a sdk.Coins object or List[Coin].
        """

        c = [Coin("uluna", 130000), Coin("ukrw", 12000000)]
        A = MsgSend(from_address=acc_address, to_address=acc_address, amount=c)
        B = MsgSend(from_address=acc_address, to_address=acc_address, amount=Coins(c))

        assert A == B

        C = MsgSend(from_address=acc_address, to_address=acc_address, amount=[])
        D = MsgSend(from_address=acc_address, to_address=acc_address, amount=Coins())

        assert C == D

        assert isinstance(A.amount, Coins)
        assert isinstance(C.amount, Coins)


class TestMsgMultiSend:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgMultiSend]:
            assert_serdes_consistent(MsgMultiSend, m)
            assert_serdes_exact(MsgMultiSend, m)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgMultiSend.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgMultiSend, m)

    def test_matches_meta(self):
        assert MsgMultiSend.type == "bank/MsgMultiSend"
        assert MsgMultiSend.action == "multisend"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, dict_list, other):
        """MsgMultiSend should validate inputs[x].address and outputs[x].address if inputs and outputs
        are passed in as dicts.
        """
        assume(not is_acc_address(other))

        bad = [{"address": acc_address, "coins": []}, {"address": other, "coins": []}]
        with pytest.raises(InvalidAccAddress):
            MsgMultiSend(inputs=bad, outputs=dict_list)
        with pytest.raises(InvalidAccAddress):
            MsgMultiSend(inputs=dict_list, outputs=bad)
        with pytest.raises(InvalidAccAddress):
            MsgMultiSend(inputs=bad, outputs=bad)

        A = MsgMultiSend(inputs=dict_list, outputs=dict_list)

    def test_constructor_takes_io_or_list_dict(
        self, input_list_with_dict, output_list_with_dict, input_list, output_list
    ):
        """MsgMultiSend should accept either list of Input/Output objects or {address: ..., coins: Coins},
        which should be treated as interoperable.
        """
        A = MsgMultiSend(inputs=input_list_with_dict, outputs=output_list_with_dict)
        B = MsgMultiSend(inputs=input_list, outputs=output_list)

        assert A == B

    def test_constructor_canonizes_dict_to_io(
        self, input_list_with_dict, output_list_with_dict, dict_list
    ):
        """MsgMultiSend should use Input/Output internally, regardless of whether the
        `inputs` or `outputs` were passed as dicts.
        """

        A = MsgMultiSend(inputs=input_list_with_dict, outputs=output_list_with_dict)
        B = MsgMultiSend(inputs=dict_list, outputs=dict_list)

        assert all(isinstance(i, Input) for i in A.inputs)
        assert all(isinstance(o, Output) for o in A.outputs)

    def test_constructor_checks_inputs_outputs(
        self, mixed_list, input_list, output_list, dict_list,
    ):
        """MsgMultiSend should prevent instatiation if an Input is passed into the output, or
        vice-versa.
        """

        with pytest.raises(TypeError):
            MsgMultiSend(inputs=mixed_list, outputs=output_list)
        with pytest.raises(TypeError):
            MsgMultiSend(inputs=input_list, outputs=mixed_list)
        with pytest.raises(TypeError):
            MsgMultiSend(inputs=output_list, outputs=output_list)
        with pytest.raises(TypeError):
            MsgMultiSend(inputs=input_list, outputs=input_list)

        A = MsgMultiSend(inputs=input_list, outputs=output_list)
        B = MsgMultiSend(inputs=dict_list, outputs=dict_list)

        assert A == B

        assert all(isinstance(i, Input) for i in A.inputs)
        assert all(isinstance(o, Output) for o in A.outputs)
        assert all(isinstance(i, Input) for i in B.inputs)
        assert all(isinstance(o, Output) for o in B.outputs)



import pytest

from jigu.key.mnemonic import MnemonicKey
from jigu.util.validation import is_acc_address, is_val_address


def compare(k: MnemonicKey, data: dict):
    assert k.mnemonic == data["mnemonic"]
    assert k.acc_address == data["address"]
    assert k.val_address == data["val_address"]
    assert k.acc_pubkey == data["pubkey"]
    assert k.val_pubkey == data["val_pubkey"]


class TestMnemonicKeyDerivation:
    """Tests ability to derive correct public / private key pair from a mnemonic."""

    # TODO: Key derivation for account, index other than (0, 0)

    def test_from_mnemonic(self, mnemonics):
        """Test whether creating a MnemonicKey with a known mnemonic generates the correct values."""
        # GIVEN a 24-character mnemonic
        # THEN give back a correctly-populated MnemonicKey
        for mnemonic in mnemonics:
            k = MnemonicKey(mnemonic["mnemonic"])
            compare(k, mnemonic)

    def test_generate(self):
        """Test whether random Mnemonic generation works."""
        k = MnemonicKey.generate()
        assert len(k.mnemonic.split()) == 24
        assert k.mnemonic != MnemonicKey.generate().mnemonic
        assert is_acc_address(k.acc_address)
        assert is_val_address(k.val_address)



"""Jigu internal utilities for serialization to / deserialization from JSON."""

from __future__ import annotations

import abc
import json
from typing import Any, Generic, TypeVar

import fastjsonschema
from box import Box

from jigu.client.lcd.api import ApiResponse
from jigu.error import JsonSchemaValidationError
from jigu.util.pretty import PrettyPrintable

_cached_schemas = dict()

T = TypeVar("T")


class JsonDeserializable(Generic[T], metaclass=abc.ABCMeta):
    """Abstract base class for objects which define a unmarshalling strategy after being
    deserialized into a Python data type (dict, list, str, etc.) from a JSON string. Can
    be used as a generic type like `JsonDeserializable[dict]` to define the Python data
    type from which the object is reconstructed."""

    __schema__ = {}

    @property
    @abc.abstractmethod
    def __schema__(self) -> dict:
        """Draft-7 compliant JSONSchema for validating inputs expressed as a Python `dict`."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_data(cls, data: T, *args, **kwargs) -> JsonDeserializable[T]:
        """Deserialize the object from a native Python format

        :param data: Python data object (dict, int, list, etc.)
        :return: unmarshalled object
        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, data: T, *args, **kwargs) -> JsonDeserializable[T]:
        """Applies JSON Schema-validation checks before attempting to recreate a
        Python object. Calls `from_data()` internally.

        :param data: Python data to construct object from
        :return: unmarshalled object
        :raises: JsonSchemaValidationError: did not pass the schema-validation check
        """
        if cls not in _cached_schemas:
            _cached_schemas[cls] = fastjsonschema.compile(cls.__schema__)
        try:
            if isinstance(data, ApiResponse):
                data = data.__result__
            _cached_schemas[cls](data)  # validate
        except fastjsonschema.JsonSchemaException as e:
            raise JsonSchemaValidationError(
                cls, data, e.message, e.value, e.name, e.definition, e.rule
            )
        return cls.from_data(data, *args, **kwargs)


def serialize_to_json(item: Any, sort: bool = False, debug: bool = False) -> str:
    """Serializes an object using the serialization strategy for `JsonSerializable`.

    :param item: object to serialize
    :type item: Any
    :param sort: sort keys alphabetically
    :type sort: bool
    :param debug: pretty-print indentation
    :type debug: bool
    """
    return json.dumps(
        item,
        indent=2 if debug else None,
        sort_keys=sort,
        separators=(",", ":") if not debug else None,
        cls=JiguJsonEncoder,
    )


class JsonSerializable(PrettyPrintable, Generic[T]):
    """Abstract base class for an object that can be serialized to a JSON string. It should
    define in `.to_data()` how to marshal its contents into a native Python data type whose]
    contents are also normally JSON-serializable, or other instances of `JsonSerializable`.
    Can be used as a type like `JsonSerializable[dict]` or `JsonSerializable[int]` to indicate
    its marshalled data value type immediately prior to conversion to a string.
    """

    def to_data(self) -> T:
        """Override this to define a marshalling strategy. By default, uses a copy of
        the object's `__dict__` property.

        :return: marshalled contents of object
        """
        return dict(self.__dict__)  # copy instead of changing object

    def to_json(self, sort: bool = False, debug: bool = False) -> str:
        """Applies the JSON-serialization strategy to the marshalled contents of the object.

        :param sort: sort by key
        :type sort: bool

        :param debug: pretty-print with indentation
        :type debug: bool

        :return: serialized object as a JSON string
        :rtype: str
        """
        return serialize_to_json(self.to_data(), sort=sort, debug=debug)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, JsonSerializable):
            return self.to_data() == other.to_data()
        else:
            return self.to_data() == other


class JiguJsonEncoder(json.JSONEncoder):
    """Encoder class for `JsonSerializable`"""

    def default(self, o) -> Any:
        if hasattr(o, "to_data"):
            return o.to_data()
        else:
            return json.JSONEncoder.default(self, o)


K = TypeVar("K")  # key
V = TypeVar("V")  # value


class JiguBox(JsonSerializable[dict], Box, Generic[K, V]):
    """JSON-Serializable version of `Box`."""

    def __init__(self, *args, **kwargs):
        Box.__init__(self, *args, **kwargs)

    def __getitem__(self, item: K, _ignore_default: bool = False) -> V:
        return Box.__getitem__(self, item, _ignore_default)

    def __repr__(self) -> str:
        return f"JiguBox{self.to_dict()!r}"

    def to_data(self):
        return self.to_dict()

    @property
    def pretty_header(self):
        return None

    @property
    def pretty_data(self):
        d = self.to_data()
        return list(d.items())



from __future__ import annotations

from contextlib import contextmanager
from typing import List, Optional, Union

import jigu.client.terra
from jigu.client.object_query import AccountQuery
from jigu.core import StdFee, StdMsg, StdSignMsg, StdTx
from jigu.key import Key

__all__ = ["Wallet"]


class Wallet(AccountQuery):
    """A Wallet is an augmented AccountQuery, and provides `chain_id`, `account_number`,
    and `sequence` information for signing transactions, which is performed by a Key."""

    def __init__(self, terra: jigu.client.terra.Terra, key: Key):
        AccountQuery.__init__(self, terra, key.acc_address)
        self.key = key
        self.terra = terra
        self._account_number = 0
        self._manual_sequence = False
        self._sequence = None

    def __repr__(self) -> str:
        return f"Wallet<{self.address}> -> {self.terra}"

    @property
    def account_number(self) -> int:
        """Account number is fetched for the first time it is found, then saved."""
        if self._account_number == 0:
            self._account_number = self.info().account_number
        return self._account_number

    @property
    def sequence(self) -> int:
        """Property that dynamically fetches sequence number everytime it is called."""
        return self.info().sequence

    def create_tx(
        self, *msgs: StdMsg, fee: Optional[StdFee] = None, memo: str = "",
    ) -> StdSignMsg:
        """Creates a sign message (`StdSignMsg`), which contains the necessary info to
        sign the transaction. Helpful to think of it as "create unsigned tx".
        """
        if not fee:
            # estimate our fee if fee not supplied
            tx = StdTx(msg=msgs, memo=memo)
            fee = self.terra.tx.estimate_fee(tx)
        if self._manual_sequence:
            sequence = self._sequence
            self._sequence += 1
        else:
            sequence = self.sequence
        return StdSignMsg(
            chain_id=self.terra.chain_id,
            account_number=self.account_number,
            sequence=sequence,
            fee=fee,
            msgs=msgs,
            memo=memo,
        )

    def sign_tx(self, *args, **kwargs):
        """Uses the Wallet's key to sign the transaction."""
        return self.key.sign_tx(*args, **kwargs)

    def create_and_sign_tx(
        self, *msgs: StdMsg, fee: Optional[StdFee] = None, memo: str = "",
    ) -> StdTx:
        """Creates a sign message, signs it, and produces a transaction in one go.
        Outputs a ready-to-broadcast `StdTx`.
        """
        return self.sign_tx(self.create_tx(*msgs, fee=fee, memo=memo))

    @contextmanager
    def manual(self) -> Wallet:
        """Manual mode is a context that creates sign messages (and transactions) with the
        sequence number incremented on each TX generation rather than polling from the
        blockchain every time, so they can be saved and broadcasted at a later time.
        """
        self._sequence = self.sequence
        self._manual_sequence = True
        yield self
        self._manual_sequence = False

    def broadcast(self, *args, **kwargs):
        return self.terra.tx.broadcast(*args, **kwargs)



import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.msg.gov import MsgDeposit, MsgSubmitProposal, MsgVote
from jigu.core.sdk import Coin, Coins
from jigu.error import InvalidAccAddress
from jigu.util.validation import is_acc_address
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the schema validation set for gov messages."""
    return load_msg_examples(tdd, msg_types=[MsgDeposit, MsgSubmitProposal, MsgVote],)


class TestMsgDeposit:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgDeposit]:
            assert_serdes_consistent(MsgDeposit, m)
            assert_serdes_exact(MsgDeposit, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgDeposit.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgDeposit, m)

    def test_matches_meta(self):
        assert MsgDeposit.type == "gov/MsgDeposit"
        assert MsgDeposit.action == "deposit"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, other):
        """MsgDeposit should validate `depositor` to be AccAddress.
        """
        assume(not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgDeposit(proposal_id=1, depositor=other, amount=[])
        A = MsgDeposit(proposal_id=1, depositor=acc_address, amount=[])
        assert A.depositor == acc_address

    def test_constructor_canonizes_amount_to_coins(self, acc_address):
        """MsgDeposit should convert List[Coins] to sdk.Coins
        """
        amnt = [Coin("ukrw", 1000)]
        A = MsgDeposit(proposal_id=1, depositor=acc_address, amount=amnt)

        assert isinstance(A.amount, Coins)
        assert A.amount == amnt


class TestMsgSubmitProposal:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgSubmitProposal]:
            assert_serdes_consistent(MsgSubmitProposal, m)
            # assert_serdes_exact(MsgSubmitProposal, m)
            # we can't have exact because serialize_to_json doesn't preserve key order.

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgSubmitProposal.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgSubmitProposal, m)

    def test_matches_meta(self):
        assert MsgSubmitProposal.type == "gov/MsgSubmitProposal"
        assert MsgSubmitProposal.action == "submit_proposal"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, other):
        """MsgSubmitProposal should validate `proposer` to be AccAddress.
        """
        assume(not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgSubmitProposal(content=None, initial_deposit=[], proposer=other)
        A = MsgSubmitProposal(content=None, initial_deposit=[], proposer=acc_address)
        assert A.proposer == acc_address

    def test_constructor_canonizes_initial_deposit_to_coins(self, acc_address):
        """MsgSubmitProposal should convert initial_deposit List[Coins] to sdk.Coins
        """
        amnt = [Coin("ukrw", 1000)]
        A = MsgSubmitProposal(content=None, initial_deposit=amnt, proposer=acc_address)

        assert isinstance(A.initial_deposit, Coins)
        assert A.initial_deposit == amnt


class TestMsgVote:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgVote]:
            assert_serdes_consistent(MsgVote, m)
            assert_serdes_exact(MsgVote, m)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgVote.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgVote, m)

    def test_matches_meta(self):
        assert MsgVote.type == "gov/MsgVote"
        assert MsgVote.action == "vote"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, other):
        """MsgVote should validate `voter` to be AccAddress."""
        assume(not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgVote(proposal_id=1, voter=other, option="")
        A = MsgVote(proposal_id=1, voter=acc_address, option="")
        assert A.voter == acc_address



from typing import Dict, Optional, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import AccAddress, Coin, Coins, Dec, ValAddress
from jigu.util.serdes import JiguBox
from jigu.util.validation import validate_acc_address, validate_val_address

__all__ = ["DistributionApi"]


class DistributionApi(BaseApi):
    def rewards_for(self, delegator: AccAddress) -> Union[ApiResponse, dict]:
        """Get an account's delegation rewards."""
        delegator = validate_acc_address(delegator)
        res = self._api_get(f"/distribution/delegators/{delegator}/rewards")
        rewards = res["rewards"] or []
        total = Coins.from_data(res["total"])
        result = JiguBox(
            {
                "rewards": {
                    r["validator_address"]: Coins.deserialize(r["reward"])
                    for r in rewards
                },
                "total": total,
            }
        )
        return project(res, result)

    # def reward(self, delegator: AccAddress, validator: ValAddress):
    #     """Get the delegation reward between a delegator and validator."""
    #     delegator = validate_acc_address(delegator)
    #     validator = validate_val_address(validator)
    #     res =  self.rewards_for(delegator)
    #     return project(res, res["via"]["validator"])

    def withdraw_address_for(
        self, delegator: AccAddress
    ) -> Union[ApiResponse, AccAddress]:
        delegator = validate_acc_address(delegator)
        return self._api_get(f"/distribution/delegators/{delegator}/withdraw_address")

    def val_rewards_for(
        self, validator: ValAddress, key: Optional[str] = None
    ) -> Union[ApiResponse, Coins, Dict[str, Coins]]:
        validator = validate_val_address(validator)
        res = self._api_get(f"/distribution/validators/{validator}")
        rewards = JiguBox(
            {
                "self_bond": Coins.deserialize(res["self_bond_rewards"]),
                "commission": Coins.deserialize(res["val_commission"]),
            }
        )
        return project(res, rewards[key] if key else rewards)

    def community_pool(
        self, denom: Optional[str] = None
    ) -> Union[ApiResponse, Coin, Coins]:
        res = self._api_get("/distribution/community_pool")
        cp = Coins.deserialize(res)
        return project(res, cp[denom] if denom else cp)

    def params(self, key: Optional[str] = None) -> Union[ApiResponse, dict, Dec, bool]:
        res = self._api_get("/distribution/parameters")
        p = res
        p["community_tax"] = Dec(p["community_tax"])
        p["base_proposer_reward"] = Dec(p["base_proposer_reward"])
        p["bonus_proposer_reward"] = Dec(p["bonus_proposer_reward"])
        return project(res, p[key] if key else JiguBox(p))



import pytest
from hypothesis import assume, given
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core import Coin, Dec
from jigu.core.denoms import uKRW, uLuna
from jigu.error import DenomIncompatibleError
from testtools import assert_serdes_consistent, assert_serdes_exact


@pytest.fixture
def d1(scope="session"):
    return uLuna


@pytest.fixture
def d2(scope="session"):
    return uKRW


coin_amount = st.one_of(st.integers(), from_schema(Dec.__schema__).map(Dec))


@pytest.mark.sdk
class TestCoinSerdes:
    @pytest.mark.serdes
    @given(c=from_schema(Coin.__schema__))
    def test_serdes_consistent(self, c):
        assert_serdes_consistent(Coin, c)

    def test_uses_correct_internal_amount(self, d1):
        examples = [
            (500, int),
            ("4.2", Dec),
            ("0", int),
            ("0.0", Dec),
            (-100000000000, int),
            ("-0.00000000000001", Dec),
        ]
        for x in examples:
            coin = Coin(d1, x[0])
            assert isinstance(coin.amount, x[1])


@pytest.mark.sdk
class TestCoinOperations:
    def test_make_coin(self, d1):
        coin = Coin(d1, "13929")
        assert coin.denom == d1
        assert coin.amount == 13929

        coin = Coin(d1, "0.006250000000000000")
        assert coin.denom == d1
        assert coin.amount == Dec("0.00625")

    @given(x=st.integers())
    def test_coin_amount_type(self, x, d1):
        # should not be different with integers
        coin = Coin(d1, x)
        coin2 = Coin(d1, str(x))
        assert coin.amount == coin2.amount

    @given(x=coin_amount, y=coin_amount)
    def test_coin_eq(self, x, y, d1, d2):
        """Tests Coin equality against another Coin."""
        assume(x != y)

        coin = Coin(d1, x)
        coin2 = Coin(d1, x)
        assert coin == coin2

        # denoms different
        coin3 = Coin(d1, x)
        coin4 = Coin(d2, x)
        assert coin3 != coin4

        # amount different
        coin5 = Coin(d1, x)
        coin6 = Coin(d1, y)
        assert coin5 != coin6

        # amount type different
        coin7 = Coin(d1, 4)
        coin8 = Coin(d1, "4.2")
        coin9 = Coin(d1, 4.2)
        assert coin7 != coin8
        assert coin8 == coin9

    @given(x=coin_amount, y=coin_amount)
    def test_coin_add(self, d1, d2, x, y):
        coin = Coin(d1, x)
        coin2 = Coin(d1, y)
        coin3 = Coin(d1, x + y)
        coinsum = coin + coin2

        assert isinstance(coinsum.amount, int) or isinstance(coinsum.amount, Dec)
        assert coin3 == coinsum
        with pytest.raises(DenomIncompatibleError):
            Coin(d1, x) + Coin(d2, x)

    @given(x=coin_amount, y=coin_amount)
    def test_coin_sub(self, d1, d2, x, y):
        coin = Coin(d1, x)
        coin2 = Coin(d1, y)
        coin3 = Coin(d1, x - y)
        coindiff = coin - coin2

        assert isinstance(coindiff.amount, int) or isinstance(coindiff.amount, Dec)
        assert coin3 == coindiff
        with pytest.raises(DenomIncompatibleError):
            Coin(d1, x) - Coin(d2, x)

    @given(x=coin_amount, y=coin_amount)
    def test_coin_compare(self, d1, d2, x, y):
        assume(x < y)
        coin = Coin(d1, x)
        coin2 = Coin(d1, y)
        coin3 = Coin(d2, x)

        coin_dup = Coin(d1, x)
        assert coin < coin2
        assert coin2 > coin

        assert coin >= coin_dup
        assert coin <= coin_dup

        with pytest.raises(DenomIncompatibleError):
            coin < coin3

    # TODO: write tests for mul, div, floordiv



from __future__ import annotations

import concurrent.futures
import sys
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List
from urllib.parse import urljoin

import requests
from requests_futures.sessions import FuturesSession

from jigu.__version__ import __version__

from .lcdrequest import LcdRequest
from .middlewares import (
    handle_codespace_errors,
    handle_general_errors,
    set_query,
    set_timeout
)

pv = sys.version_info

CLIENT_HEADERS = {
    "Accept": "application/json",
    "User-Agent": f"jigu v{__version__} (Python {pv[0]}.{pv[1]}.{pv[2]})",
}


def create_session():
    session = requests.session()
    session.headers.update(CLIENT_HEADERS)
    return session


def create_futures_session():
    session = FuturesSession()
    session.headers.update(CLIENT_HEADERS)
    return session


class LcdClient(object):
    """Manages the connection to a Terra LCD node, helps with procesing requests."""

    def __init__(self, terra, url: str = "", timeout: int = 30, threads: int = 1):
        self.url = url
        self.timeout = timeout
        self._nthreads = threads
        self.executor = ThreadPoolExecutor(max_workers=self._nthreads)
        self.session = create_session()
        self.futures_session = create_futures_session()
        self.request_middlewares = [set_timeout, set_query]
        self.response_middlewares = [handle_codespace_errors, handle_general_errors]

    @property
    def threads(self) -> int:
        return self._nthreads

    @threads.setter
    def threads(self, other: int):
        self._nthreads = other
        self.executor = ThreadPoolExecutor(max_workers=self._nthreads)

    def _create_url(self, path):
        return urljoin(self.url, path)

    def _build_request(self, method, path, kwargs):
        """Applies middlewares to request."""
        request = LcdRequest(method=method, url=self._create_url(path), kwargs=kwargs)
        for req_m in self.request_middlewares:
            request = req_m(self, request)  # TODO: use reduce()
        return request

    def _send_request(self, request, use_future=False):
        session = self.futures_session if use_future else self.session
        return getattr(session, request.method)(request.url, **request.kwargs)

    def handle_response(self, response):
        """Applies middlewares to response."""
        r = response
        for resp_m in self.response_middlewares:
            r = resp_m(self, r)
        return r

    def _request(self, method, path, **kwargs):
        request = self._build_request(method, path, kwargs)
        response = self._send_request(request)
        return self.handle_response(response)

    def get(self, path, build=False, **kwargs):
        if build:  # builds a GET request, but does not send.
            return self._build_request("get", path, kwargs)
        return self._request("get", path, **kwargs)

    def get_threaded(self, paths, **kwargs):
        reqs: List[LcdRequest] = [
            self.get(path, build=True, **kwargs) for path in paths
        ]
        futures = {}
        for i, req in enumerate(reqs):
            futures[self._send_request(req, use_future=True)] = i
        responses = [None] * len(futures)
        resp = concurrent.futures.as_completed(futures)
        for r in resp:
            responses[futures[r]] = self.handle_response(r.result())
        return responses

    def post(self, path, build=False, **kwargs):
        if build:
            return self._build_request("post", path, kwargs)
        return self._request("post", path, **kwargs)



import pytest

from testtools import LcdRequestTest


class TestSupplyApi:
    def test_supply_request(self, mock_terra):
        try:
            mock_terra.supply()
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/supply/total"



import json
from typing import Dict, Iterable, List

import wrapt
from toolz import groupby, valmap

from jigu.core import StdMsg
from jigu.query.event import EventsQuery
from jigu.util.pretty import PrettyPrintable, see
from jigu.util.serdes import JiguBox
from jigu.util.validation import Schemas as S

__all__ = ["MsgInfo", "MsgInfosQuery"]


class MsgInfo(wrapt.ObjectProxy):

    __schema__ = S.OBJECT(
        msg_index=S.INTEGER,
        success=S.BOOLEAN,
        log=S.STRING,
        events=S.ARRAY(
            S.OBJECT(
                type=S.STRING,
                attributes=S.ARRAY(S.OBJECT(key=S.STRING, value=S.STRING)),
            )
        ),
    )

    def __init__(self, msg: StdMsg, success: bool, log: dict, events: EventsQuery):
        wrapt.ObjectProxy.__init__(self, msg)
        try:
            log = json.loads(log)
            if type(log) == dict:
                log = JiguBox(log)
        except:
            log = None
        self._self_success = success
        self._self_log = log
        self._self_events = events
        self._self_pretty_data = None

    @property
    def success(self):
        return self._self_success

    @property
    def log(self):
        return self._self_log

    @property
    def events(self):
        return self._self_events

    def __eq__(self, other):
        return (
            isinstance(other, MsgInfo)
            and self.success == other.success
            and self.log == other.log
            and self.events == other.events
        )

    def __ne__(self, other):
        # we have to do this because objectproxy uses the underlying wrapped's __neq__
        return not self == other

    @property
    def pretty_data(self):
        d = dict(self.__dict__)
        items = list(d.items())
        items.append(("success", self.success))
        if self.log:
            items.append(("log", self.log))
        items.append(("events", self.events))
        return items

    def _pretty_output(self, path: str = ""):
        return PrettyPrintable._pretty_output(self, path)

    def _pretty_repr_(self, path: str = "") -> str:
        return PrettyPrintable._pretty_repr_(self, path)

    @property
    def _pp(self):
        """Shortcut for seeing pretty-printing output."""
        see(self)
        return None


class MsgInfosQuery(PrettyPrintable):
    """Convenience query structure around a list of MsgInfo objects.

    Use the variable name "msgs" if possible.

    msgs[n] (n = int) --> n-th MsgInfo
    msgs["msg-type / msg-action"] --> recursive query of MsgInfos of that type / action
    msgs.msg_action --> list of msginfos of that action
    for e in msgs: --> iterate over msginfos
    "send" in msgs: --> whether msgs of this action are in collection
    """

    def __init__(self, msginfos: Iterable[MsgInfo]):
        self.msginfos = list(msginfos)

    def __repr__(self) -> str:
        r = ", ".join(f"{t}: {n}" for t, n in self.actions.items())
        if r == "":
            r = "(empty)"
        return f"<MsgInfosQuery {r}>"

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.msginfos[key]
        if isinstance(key, slice):
            return MsgInfosQuery(self.msginfos[key])
        elif isinstance(key, str):
            return MsgInfosQuery(
                [m for m in self.msginfos if m.type == key or m.action == key]
            )
        elif hasattr(key, "type") and hasattr(key, "action"):
            return MsgInfosQuery(
                [
                    m
                    for m in self.msginfos
                    if m.type == key.type and m.action == key.action
                ]
            )
        else:
            return self.msginfos[key]

    def __getattr__(self, name):
        return self[name]

    def __iter__(self):
        return iter(self.msginfos)

    def __contains__(self, item):
        if isinstance(item, StdMsg):
            return item.type in self.types
        else:
            return item in self.types or self.actions

    @property
    def logs(self) -> List[JiguBox]:
        msg_logs = [m.log for m in self.msginfos if m.log is not None]
        return msg_logs

    @property
    def events(self) -> EventsQuery:
        msg_events = [event for m in self.msginfos for event in m.events]
        return EventsQuery(msg_events)

    @property
    def types(self) -> Dict[str, int]:
        msg_types = groupby(lambda m: m.type, self.msginfos)
        msg_types = valmap(len, msg_types)
        return msg_types

    @property
    def actions(self) -> Dict[str, int]:
        msg_actions = groupby(lambda m: m.action, self.msginfos)
        msg_actions = valmap(len, msg_actions)
        return msg_actions

    def __len__(self) -> int:
        return len(self.msginfos)

    # TODO: add printing mechanism?
    # Example:
    # a = JiguBox({ x: MsgInfo(...) })
    # print(a)

    @property
    def pretty_data(self):
        items = groupby(lambda m: str(m.__class__.__name__), self.msginfos)
        items = valmap(len, items)
        return items.items()



from typing import Optional, Union

from jigu.client.lcd.api import BaseApi, project
from jigu.core import Dec, Timestamp, ValConsPubKey
from jigu.util.serdes import JiguBox
from jigu.util.validation import validate_val_consaddress, validate_val_conspubkey

__all__ = ["SlashingApi"]


def to_signing_info(si):
    box = JiguBox(
        si,
        box_recast={
            "start_height": int,
            "index_offset": int,
            # "jailed_until": Timestamp,
            "missed_blocks_counter": int,
        },
    )
    box.jailed_until = Timestamp.from_data(box.jailed_until)
    return box


# TODO: document and revamp
class SlashingApi(BaseApi):
    def signing_info_by_pubkey(self, pubkey: ValConsPubKey):
        validate_val_conspubkey(pubkey)
        res = self._api_get(f"/slashing/validators/{pubkey}/signing_info")
        return project(res, to_signing_info(res))

    def signing_info_by_address(self, address: str = None):
        if address:
            validate_val_consaddress(address)
        res = self._api_get("/slashing/signing_infos")
        infos = [to_signing_info(r) for r in res]
        by_address = JiguBox({info.address: info for info in infos})
        return project(res, by_address[address] if address else by_address)

    def params(self, key: Optional[str] = None) -> Union[int, Dec, dict]:
        res = self._api_get("/slashing/parameters")
        p = JiguBox(
            res,
            box_recast={
                "max_evidence_age": int,
                "signed_blocks_window": int,
                "min_signed_per_window": Dec,
                "downtime_jail_duration": int,
                "slash_fraction_double_sign": Dec,
                "slash_fraction_downtime": Dec,
            },
        )
        return project(res, p[key] if key else p)



"""Block transformers.

A function that is called after the block has been fetched and deserialized,
to transform the block data and add more info, such as turning Amino-encoded
TXs from TxInfo's __rawtx__ (if it exists) into full TxInfo data.

The default behavior is to fetch all the transactions by their hash, because
we do not yet have the means to decode Amino from Python. Having one would
immensely boost performance for block-related queries. Feel free to disable
this automatic behavior by setting your Terra instance to not fetch transactions.

(see `fetch_txs_transformer`)

You can also swap out the `transform_block` function with your own implementation,
    which should implement the signature:

    def my_transformer(terra: jigu.Terra, block: jigu.core.Block) -> jigu.core.Block

Swapping out this functionality:

>>> terra.blocks.transformer = my_transformer

And voila! You should get your transform block as the result of any queries!
"""

from jigu.core import Block
# TODO: add more efficient block transformer by adding go-amino bindings for Python
from jigu.query.txinfo import TxInfosQuery
from jigu.util import hash_amino

__all__ = ["FetchTxInfoBlockTransformer"]


class FetchTxInfoBlockTransformer:
    """Fetches TxInfo from data in `block.txs`, which it expects to contain Amino-encoded
    TX data."""

    def __init__(self, terra):
        self.terra = terra  # need terra instance to fetch txinfo

    def __call__(self, block: Block) -> Block:
        txhashes = [hash_amino(txdata) for txdata in block.txs]
        txs = self.terra.tx._tx_info_threaded(txhashes)
        block.txs = TxInfosQuery(txs)
        return block



"""Test suite translated from https://github.com/terra-project/terra-js/blob/master/tests/utils/txUtils.test.ts"""

import pytest

from jigu import Terra
from jigu.key.mnemonic import MnemonicKey
from jigu.core import (
    Coin,
    Coins,
    StdSignMsg,
    StdTx,
    StdFee,
    StdSignature,
    Input,
    Output,
)
from jigu.core.msg import MsgSend, MsgMultiSend, MsgVote
from jigu.util import hash_amino


@pytest.fixture
def terra():
    return Terra("columbus-3", "https://lcd.terra.dev")


@pytest.fixture
def master_key():
    return MnemonicKey(
        "island relax shop such yellow opinion find know caught erode blue dolphin behind coach tattoo light focus snake common size analyst imitate employ walnut"
    )


@pytest.fixture
def master_key2():
    return MnemonicKey(
        "spatial fantasy weekend romance entire million celery final moon solid route theory way hockey north trigger advice balcony melody fabric alter bullet twice push"
    )


@pytest.fixture
def a1():
    return MnemonicKey(
        "swamp increase solar renew twelve easily possible pig ostrich harvest more indicate lion denial kind target small dumb mercy under proud arrive gentle field"
    )


@pytest.fixture
def a2():
    return MnemonicKey(
        "service frozen keen unveil luggage initial surge name conduct mesh soup escape weather gas clown brand holiday result protect chat plug false pitch little"
    )


@pytest.fixture
def a3():
    return MnemonicKey(
        "corn peasant blue sight spy three stove confirm night brother vote dish reduce sick observe outside vacant arena laugh devote exotic wasp supply rally"
    )


def test_sign_tx(master_key):
    unsigned = StdSignMsg(
        msgs=[
            MsgSend(
                from_address=master_key.acc_address,
                to_address="terra1wg2mlrxdmnnkkykgqg4znky86nyrtc45q336yv",
                amount=[Coin("uluna", 100_000_000)],
            )
        ],
        sequence=0,
        account_number=45,
        chain_id="columbus-3-testnet",
        fee=StdFee(gas=46_467, amount=[Coin("uluna", 698)]),
    )
    signature = master_key.create_signature(unsigned)
    assert (
        signature.signature
        == "FJKAXRxNB5ruqukhVqZf3S/muZEUmZD10fVmWycdVIxVWiCXXFsUy2VY2jINEOUGNwfrqEZsT2dUfAvWj8obLg=="
    )


def test_multisig(a1, a2, a3):
    receiver = "terra1ptdx6akgk7wwemlk5j73artt5t6j8am08ql3qv"
    multisig = {
        "address": "terra16ddrexknvk2e443jsnle4n6s2ewjc6z3mjcu6d",
        "account_number": 46,
        "sequence": 0,
    }
    assert a1.acc_address == "terra12dazwl3yq6nwrce052ah3fudkarglsgvacyvl9"
    assert a2.acc_address == "terra1jqw25580qljucyy2xkpp7j02kd4mwx69wvfgf9"
    assert a3.acc_address == "terra13hrg8ul0p7sh85jwalh3leysdrw9swh44dql2h"

    send = MsgSend(
        from_address=multisig["address"],
        to_address=receiver,
        amount=[Coin("uluna", 100_000_000)],
    )

    unsigned = StdSignMsg(
        chain_id="columbus-3-testnet",
        account_number=multisig["account_number"],
        sequence=multisig["sequence"],
        msgs=[send],
        fee=StdFee.make(gas=50_000, uluna=750),
    )
    a1_sig = a1.create_signature(unsigned)
    assert (
        a1_sig.signature
        == "/kIFqGnmgOqMzf7guoe1eDTA1W5TjJcelJSRBdN0CTRyyxTMIbsxd+wL4fatHAq4hYOTf/zxD4l5xyU7/POZyg=="
    )
    a2_sig = a2.create_signature(unsigned)
    assert (
        a2_sig.signature
        == "hEjv9CnXQa89robHVsHS3GDZJiunnNb8xqziWD8D4aAuBXwxDzUXY14IE7q9Z3Qh0VMb3FBHuogHi7QZn2pM9g=="
    )
    a3_sig = a3.create_signature(unsigned)
    assert (
        a3_sig.signature
        == "CwHdmwC9ADtr5cTUdRZEfAcA8d1bgkF8fB+DcbB6MBB6amJz51WQYfVE1VgVTEY8Lyzg8+s8gX6nkqkXPeX72A=="
    )


def test_txid(terra):

    signature = StdSignature.from_data(
        {
            "signature": "+SnQyRQZ536m0VLTwWFn6WTlmV0ZP+EI08lIGbZFhvYMLPA+Dld3qaTFKwgJEd7kZrAb5OPWBUhiOc9326daEw==",
            "pub_key": {
                "type": "tendermint/PubKeySecp256k1",
                "value": "Ar+guke5UuM2XEZ9/ouPhAQbYs+f7y6jQCtGlI2lj1ZH",
            },
        }
    )

    # manually build TX
    tx = StdTx(
        msg=[
            MsgSend(
                from_address="terra1wg2mlrxdmnnkkykgqg4znky86nyrtc45q336yv",
                to_address="terra18h5pmhrz45z2ne7lz4nfd7cdfwl3jfeu99e7a5",
                amount=[Coin("uluna", 100_000_000)],
            )
        ],
        fee=StdFee.make(54_260, ukrw=814),
        signatures=[signature],
    )
    # get amino encoding
    amino = terra.tx.encode(tx)
    txhash = hash_amino(amino)
    assert txhash == "028b2acc80d244241114bf20b2982889201eca42bd400bc8f3a9d2162b5f0a3e"


def test_multisend(master_key2):
    """This test highlights how Jigu is flexible to accomodate many input styles."""

    multisend = MsgMultiSend(
        inputs=[
            Input(
                address=master_key2.acc_address,
                coins=[Coin("uluna", 1_000_000), Coin(denom="usdr", amount="1000000")],
            )
        ],
        outputs=[
            Output(
                "terra12dazwl3yq6nwrce052ah3fudkarglsgvacyvl9", [Coin("uluna", 500000)]
            ),
            {
                "address": "terra1ptdx6akgk7wwemlk5j73artt5t6j8am08ql3qv",
                "coins": Coins() + Coin("uluna", 500000) + Coin("usdr", 1000000),
            },
        ],
    )

    unsigned = StdSignMsg(
        account_number=47,
        sequence=0,
        chain_id="columbus-3-testnet",
        msgs=[multisend],
        fee=StdFee.make(100_000, uluna=1500, usdr=1000),
        memo="1234",
    )

    tx = master_key2.sign_tx(unsigned)

    assert (
        tx.signatures[0].signature
        == "YA/ToXLxuuAOQlpm5trbIUu2zv5NfBmeHz2jmXgNrt8jP+odukerfri3DUXAJuhETAMHVVV78t7Q4xC0j+CVkA=="
    )


def test_vote(terra):
    vote = MsgVote(
        proposal_id=1,
        voter="terra1wg2mlrxdmnnkkykgqg4znky86nyrtc45q336yv",
        option=MsgVote.YES,
    )

    signature = StdSignature.from_data(
        {
            "signature": "Xn93OEY/HYXxE2rqL6O5Cj0r/D+GbaPxEgvKM+hMFQJA/95RE4dLPJdLP5HpmeLrBm46GI0V6w27pR1Pq4mH+g==",
            "pub_key": {
                "type": "tendermint/PubKeySecp256k1",
                "value": "Ar+guke5UuM2XEZ9/ouPhAQbYs+f7y6jQCtGlI2lj1ZH",
            },
        }
    )

    tx = StdTx(msg=[vote], fee=StdFee.make(70_000, uluna=1050), signatures=[signature])

    amino = terra.tx.encode(tx)
    txhash = hash_amino(amino)

    assert txhash == "f1af6a4193c500578a457dd68ee34131e15d8e1c462bba647c43e882fcd74e57"



from __future__ import annotations

from typing import Any, Optional, Union

import requests
import wrapt

from jigu.error import UnhandledResponse

__all__ = ["ApiResponse", "BaseApi", "project"]


class ApiResponse(wrapt.ObjectProxy):
    """Wraps the LCD API response in an object proxy that keeps track of the chain height
    reported at the time of the request, as well as the orginal requests.Response (which
    includes the request) object. This lets users use the parsed response directly, as
    well as keep potentitally useful information."""

    def __init__(
        self, result: Any, response: requests.Response, height: Optional[int] = None
    ):
        wrapt.ObjectProxy.__init__(self, result)
        self._self_result = result
        self._self_response = response
        self._self_height = height

    def __repr__(self):
        return f"<ApiResponse {self.__wrapped__!r}>"

    @property
    def __type__(self):
        return type(self._self_result)

    @property
    def __result__(self) -> Any:
        """Unwraps the APIResponse to get the underlying wrapped object."""
        return self._self_result

    @property
    def __request__(self) -> requests.Request:
        """The `requests.Requests` object used for the request."""
        return self._self_response.request

    @property
    def __response__(self) -> requests.Response:
        """The original `requests.Response` object."""
        return self._self_response

    @property
    def __height__(self) -> int:
        """Chain height reported at the response (may be `None` for some queries)."""
        return self._self_height


def project(a: ApiResponse, b: Any) -> Union[ApiResponse, Any]:
    """Creates a new `ApiResponse` out of `b` by projecting `a`'s response and height
    metadata."""
    # we don't want russian dolls of ApiResponses
    while isinstance(b, ApiResponse):
        b = b.__result__  # unpack
    return ApiResponse(b, a.__response__, a.__height__)


class BaseApi(object):
    """Base class for Terra API objects. Provides functions for turning an API response
    into a wrapped object proxy called an `ApiResponse`."""

    def __init__(self, terra):
        self.terra = terra

    @property
    def lcd(self):
        return self.terra.lcd

    def _api_get(self, path, unwrap=True, **kwargs) -> Union[ApiResponse, Any]:
        return self._handle_response(self.terra.lcd.get(path, **kwargs), unwrap)

    def _api_post(self, path, unwrap=True, **kwargs) -> Union[ApiResponse, Any]:
        return self._handle_response(self.terra.lcd.post(path, **kwargs), unwrap)

    @staticmethod
    def _handle_response(
        response: requests.Response, unwrap=True
    ) -> Union[ApiResponse, Any]:
        """Creates an ApiResponse object from the LCD Client's response.

        :param response: LCD response
        :param unwrap: whether the LCD response will include height
        """
        # TODO: refactor into middleware? --
        # NOTE: This actually should be in the last step, but the user can decide to do
        # some post-processing on their own, such as parsing, by writing a library that
        # parses the result. Outside the scope of the SDK.
        try:
            res = response.json()
            if unwrap:
                return ApiResponse(res["result"], response, int(res["height"]))
            else:
                return ApiResponse(res, response, None)
        except ValueError:  # if the response cannot be parsed to JSON
            raise UnhandledResponse(f"unhandled response {str(response.content)}")



import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema

from jigu.core.staking.delegation import Delegation, Redelegation, UnbondingDelegation
from testtools import assert_serdes_consistent, assert_serdes_exact, load_data


@pytest.fixture(scope="module")
def delgn_examples(tdd):
    return load_data(tdd, "objects/staking/Delegation.json")


@pytest.fixture(scope="module")
def udelgn_examples(tdd):
    return load_data(tdd, "objects/staking/UnbondingDelegation.json")


@pytest.fixture(scope="module")
def rdelgn_examples(tdd):
    return load_data(tdd, "objects/staking/Redelegation.json")


class TestDelegationSerdes:
    def test_schema_valid(self, delgn_examples):
        for d in delgn_examples:
            assert_serdes_consistent(Delegation, d)
            assert_serdes_exact(Delegation, d)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(d=from_schema(Delegation.__schema__))
    def test_schema(self, d):
        assert_serdes_consistent(Delegation, d)


class TestUnbondingDelegationSerdes:
    def test_schema_valid(self, udelgn_examples):
        for d in udelgn_examples:
            assert_serdes_consistent(UnbondingDelegation, d)
            assert_serdes_exact(UnbondingDelegation, d)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(d=from_schema(UnbondingDelegation.__schema__))
    def test_schema(self, d):
        assert_serdes_consistent(UnbondingDelegation, d)


class TestRedelegationSerdes:
    def test_schema_valid(self, rdelgn_examples):
        for d in rdelgn_examples:
            assert_serdes_consistent(Redelegation, d)
            assert_serdes_exact(Redelegation, d)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(d=from_schema(Redelegation.__schema__))
    def test_schema(self, d):
        assert_serdes_consistent(Redelegation, d)



from __future__ import annotations

from bech32 import bech32_decode, bech32_encode

from jigu.util.validation import (
    validate_acc_address,
    validate_val_address,
    validate_val_consaddress
)

__all__ = ["AccAddress", "ValAddress", "ValConsAddress"]


class AccAddress(str):
    """`terra-` prefixed Bech32-enconded account address, works anywhere a `str` is
    accepted but will perform verification on the checksum.

    :param arg: string-convertable object to convert.
    """

    def __new__(cls, arg):
        arg = validate_acc_address(arg)
        return str.__new__(cls, arg)

    @property
    def val_address(self) -> ValAddress:
        """The associated validator operator address."""
        decoded = bech32_decode(self)
        return ValAddress(bech32_encode("terravaloper", decoded[1]))


class ValAddress(str):
    """`terravaloper-` prefixed Bech32-enconded account address, works anywhere a `str`
    is accepted but will perform verification on the checksum.

    :param arg: string-convertable object to convert.
    """

    def __new__(cls, arg):
        arg = validate_val_address(arg)
        return str.__new__(cls, arg)

    @property
    def acc_address(self) -> AccAddress:
        """The associated account address."""
        decoded = bech32_decode(self)
        return AccAddress(bech32_encode("terra", decoded[1]))


class ValConsAddress(str):
    """`terravalcons-` prefixed Bech32-enconded consensus address, works anywhere a `str`
    is accepted but will perform verification on the checksum.

    :param arg: string-convertable object to convert.
    """

    def __new__(cls, arg):
        validate_val_consaddress(arg)
        return str.__new__(cls, arg)



from __future__ import annotations

import requests
from fastjsonschema.exceptions import JsonSchemaException

### Base Classes


class JiguException(Exception):
    pass


class JiguWarning(Warning):
    pass


### Validation Errors


class ValidationError(JiguException):
    """Ambiguous validation check failed."""


class DenomIncompatibleError(ValidationError, ValueError):
    """An operation requiring compatible Coins encountered incompatible denoms."""


class InvalidAccAddress(ValidationError, ValueError):
    """An incorrect value was passed into a method that expects a valid Account
    address (Bech32-encoded address prefixed with terra-)."""


class InvalidValAddress(ValidationError, ValueError):
    """An incorrect value was passed into a method that expects a valid Validator
    (operator) address (Bech32-encoded address prefixed with terravaloper-)."""


class InvalidValConsAddress(ValidationError, ValueError):
    """An incorrect value was passed into a method that expects a valid Validator
    (consensus) address (Bech32-encoded pubkey prefixed with terravalcons-)."""


class InvalidAccPubKey(ValidationError, ValueError):
    """An incorrect value was passed into a method that expects a valid Account
    public key (Bech32-encoded pubkey prefixed with terrapub-)."""


class InvalidValPubKey(ValidationError, ValueError):
    """An incorrect value was passed into a method that expects a valid Validator
    public key (Bech32-encoded pubkey prefixed with terravalpub-)."""


class InvalidValConsPubKey(ValidationError, ValueError):
    """An incorrect value was passed into a method that expects a valid Validator
    Consensus (Tendermint) pubkey (Bech32-encoded pubkey prefixed with terravalconspub-)."""


class InvalidParamChange(ValidationError, ValueError):
    """The parameter change does not adhere to specified schema."""


class JsonSchemaValidationError(ValidationError, JsonSchemaException):
    """Failed attempt to deserialize an object using a :class:`JsonDeserializable`."""

    def __init__(self, cls, data, message, value, name, definition, rule):
        JsonSchemaException.__init__(self, message, value, name, definition, rule)
        self.cls = cls
        self.data = data


### Terra Node IO Errors


class JiguRequestError(JiguException):
    def __init__(
        self, status_code: int, error_msg: str, request: requests.Request = None
    ):
        Exception.__init__(self, f"({status_code}) {error_msg}")
        self.status_code = status_code
        self.message = error_msg
        self.request = request


class UnhandledResponse(JiguRequestError):
    """Jigu did not handle the response, but status code 200."""


class BadRequest(JiguRequestError):
    """Malformed request on the part of the user. (4xx errors)"""


class LcdInternalError(JiguRequestError):
    """An error occured on the server while processing the request. (5xx errors)"""


### RPC Errors


class RpcError(JiguRequestError):
    """Error at the RPC node level. Status code is RPC status code, not HTTP."""


class RpcInternalError(RpcError):
    """An internal error occured in the RPC node."""


RPC_ERRORS = {-32603: RpcInternalError}


def get_rpc_error(status_code: int, error_msg: str, request: requests.Request = None):
    try:
        rpc_error = RPC_ERRORS[status_code]
    except KeyError:
        rpc_error = RpcError  # generic fallback
    return rpc_error(status_code, error_msg, request)


### Codespace Errors
class CodespaceError(JiguException):
    """A more specific node error with information about the originating module."""

    def __init__(self, codespace: str, code: int, message: str, **kwargs):
        Exception.__init__(self, f"{codespace} (code {code}) {message}")
        self.codespace = codespace
        self.code = code
        self.message = message
        for k, v in kwargs.items():
            self.__setattr__(k, v)


#### TX Errors


class TxError(JiguException):
    """Broadcasting error due to something wrong with the transaction's content."""

    def __init__(self, message: str, tx, broadcast_result=None):
        Exception.__init__(self, message)
        self.message = message
        self.tx = tx
        self.broadcast_result = broadcast_result


class TxCodespaceError(CodespaceError, TxError):
    """Broadcasting error due to something wrong in the transaction's message content.
    (more specific and due to an error in the message, like denom not existing)."""


class InsufficientFee(TxCodespaceError, ValueError):
    """Transaction was not processed due to insufficient fees."""


class OutOfGasException(TxCodespaceError):
    """Transaction was accepted but validator ran out of gas while processing."""


# tx errors by codespace, error code
CODESPACE_ERRORS = {"sdk": {12: OutOfGasException, 14: InsufficientFee}}


def get_codespace_error(
    codespace: str, code: int, message: str, was_from_tx=False, **kwargs
):
    try:
        error_type = CODESPACE_ERRORS[codespace][code]
    except KeyError:
        error_type = (
            TxCodespaceError if was_from_tx else CodespaceError
        )  # generic fallback
    return error_type(codespace, code, message, **kwargs)


### Misc Errors


class DenomNotFound(JiguException, KeyError):
    """The requested denomination as not found."""


# Warnings


class AccountNotFoundWarning(JiguWarning):
    """Account lookup was performed but ended up with an empty account."""



import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema

from jigu.core.auth.account import Account
from testtools import assert_serdes_consistent, assert_serdes_exact, load_data


@pytest.fixture
def examples(tdd):
    return load_data(tdd, "objects/auth/Account.json")


class TestAccountSerdes:
    def test_schema_valid(self, examples):
        for acc in examples:
            assert_serdes_consistent(Account, acc)
            assert_serdes_exact(Account, acc)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(acc=from_schema(Account.__schema__))
    def test_serdes_consistent(self, acc):
        assert_serdes_consistent(Account, acc)



from __future__ import annotations

from dataclasses import dataclass

from jigu.core.proposal import Content
from jigu.util.validation import Schemas as S

__all__ = ["TextProposal"]


@dataclass
class TextProposal(Content):

    type = "gov/TextProposal"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^gov/TextProposal\Z"),
        value=S.OBJECT(title=S.STRING, description=S.STRING),
    )

    title: str
    description: str

    @classmethod
    def from_data(cls, data: dict) -> TextProposal:
        data = data["value"]
        return cls(title=data["title"], description=data["description"])



from __future__ import annotations

from dataclasses import dataclass

from jigu.core import AccAddress, Coins
from jigu.core.proposal import Content
from jigu.util.validation import Schemas as S

__all__ = ["CommunityPoolSpendProposal"]


@dataclass
class CommunityPoolSpendProposal(Content):

    type = "distribution/CommunityPoolSpendProposal"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^distribution/CommunityPoolSpendProposal\Z"),
        value=S.OBJECT(
            title=S.STRING,
            description=S.STRING,
            recipient=S.ACC_ADDRESS,
            amount=Coins.__schema__,
        ),
    )

    title: str
    description: str
    recipient: AccAddress
    amount: Coins

    @classmethod
    def from_data(cls, data: dict) -> CommunityPoolSpendProposal:
        data = data["value"]
        return cls(
            title=data["title"],
            description=data["description"],
            recipient=data["recipient"],
            amount=Coins.from_data(data["amount"]),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

from jigu.core import AccAddress, Coin, Dec, Timestamp, ValAddress
from jigu.core.denoms import uLuna
from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

__all__ = [
    "Delegation",
    "UnbondingDelegation",
    "UnbondingEntry",
    "Redelegation",
    "RedelegationEntry",
]


@dataclass
class Delegation(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        delegator_address=S.ACC_ADDRESS,
        validator_address=S.VAL_ADDRESS,
        shares=Dec.__schema__,
        balance=S.STRING_INTEGER,
    )

    delegator_address: AccAddress
    validator_address: ValAddress
    shares: Coin
    balance: Coin

    def to_data(self) -> dict:
        return {
            "delegator_address": self.delegator_address,
            "validator_address": self.validator_address,
            "shares": str(self.shares.amount),
            "balance": str(self.balance.amount),
        }

    @classmethod
    def from_data(cls, data: Dict[str, str]) -> Delegation:
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            shares=Coin(uLuna, data["shares"]),
            balance=Coin(uLuna, data["balance"]),
        )


@dataclass
class UnbondingEntry(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        initial_balance=S.STRING_INTEGER,
        balance=S.STRING_INTEGER,
        creation_height=S.STRING_INTEGER,
        completion_time=Timestamp.__schema__,
    )

    initial_balance: Coin
    balance: Coin
    creation_height: int
    completion_time: Timestamp

    def to_data(self) -> dict:
        return {
            "initial_balance": str(self.initial_balance.amount),
            "balance": str(self.balance.amount),
            "creation_height": str(self.creation_height),
            "completion_time": self.completion_time,
        }

    @classmethod
    def from_data(cls, data: Dict[str, str]) -> UnbondingEntry:
        return cls(
            initial_balance=Coin(uLuna, data["initial_balance"]),
            balance=Coin(uLuna, data["balance"]),
            creation_height=int(data["creation_height"]),
            completion_time=Timestamp.from_data(data["completion_time"]),
        )


@dataclass
class UnbondingDelegation(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        delegator_address=S.ACC_ADDRESS,
        validator_address=S.VAL_ADDRESS,
        entries=S.ARRAY(UnbondingEntry.__schema__),
    )

    delegator_address: AccAddress
    validator_address: ValAddress
    entries: List[UnbondingEntry]

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> UnbondingDelegation:
        entries = [UnbondingEntry.from_data(entry) for entry in data["entries"]]
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            entries=entries,
        )


@dataclass
class RedelegationEntry(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        initial_balance=S.STRING_INTEGER,
        balance=S.STRING_INTEGER,
        shares_dst=Dec.__schema__,
        creation_height=S.INTEGER,
        completion_time=Timestamp.__schema__,
    )

    initial_balance: Coin
    balance: Coin
    shares_dst: Coin
    creation_height: int
    completion_time: Timestamp

    def to_data(self) -> dict:
        return {
            "creation_height": self.creation_height,
            "completion_time": self.completion_time,
            "initial_balance": str(self.initial_balance.amount),
            "balance": str(self.balance.amount),
            "shares_dst": str(self.shares_dst.amount),
        }

    @classmethod
    def from_data(cls, data: dict) -> RedelegationEntry:
        return cls(
            initial_balance=Coin(uLuna, data["initial_balance"]),
            balance=Coin(uLuna, data["balance"]),
            shares_dst=Coin(uLuna, data["shares_dst"]),
            creation_height=int(data["creation_height"]),
            completion_time=Timestamp.from_data(data["completion_time"]),
        )


@dataclass
class Redelegation(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        delegator_address=S.ACC_ADDRESS,
        validator_src_address=S.VAL_ADDRESS,
        validator_dst_address=S.VAL_ADDRESS,
        entries=S.ARRAY(RedelegationEntry.__schema__),
    )

    delegator_address: AccAddress
    validator_src_address: ValAddress
    validator_dst_address: ValAddress
    entries: List[RedelegationEntry]

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> Redelegation:
        entries = [RedelegationEntry.from_data(re) for re in data["entries"]]
        return cls(
            delegator_address=data["delegator_address"],
            validator_src_address=data["validator_src_address"],
            validator_dst_address=data["validator_dst_address"],
            entries=entries,
        )



import json

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.auth.transaction import StdFee


class TestStdFeeSerdes:
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(fee=from_schema(StdFee.__schema__))
    def test_schema(self, fee):
        x = StdFee.deserialize(fee)
        y = StdFee.deserialize(json.loads(x.to_json()))
        assert x == y



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from bech32 import bech32_decode, bech32_encode

from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S
from jigu.util.validation import (
    validate_acc_pubkey,
    validate_val_conspubkey,
    validate_val_pubkey
)

__all__ = ["PublicKey", "AccPubKey", "ValPubKey", "ValConsPubKey"]


@dataclass
class PublicKey(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        type=S.STRING,
        value=S.ANY(
            S.STRING,
            S.OBJECT(
                threshold=S.STRING_INTEGER,
                pubkeys=S.ARRAY(S.OBJECT(type=S.STRING, value=S.STRING)),
            ),
        ),
    )

    type: str = "tendermint/PubKeySecp256k1"
    value: Any = None

    @classmethod
    def from_data(cls, data: dict) -> PublicKey:
        # TODO: apply None-coalescing feature to root JsonDeserializable
        # viz: JsonDeserializable
        if data is None:
            return None
        return cls(type=data["type"], value=data["value"])


class AccPubKey(str):
    """`terrapub-` prefixed Bech32-enconded account public eky, works anywhere a `str` is
    accepted but will perform verification on the checksum.

    :param arg: string-convertable object to convert.
    """

    def __new__(cls, arg):
        validate_acc_pubkey(arg)
        return str.__new__(cls, arg)

    @property
    def val_pubkey(self) -> ValPubKey:
        """The associated validator public key."""
        decoded = bech32_decode(self)
        return ValPubKey(bech32_encode("terravalpub", decoded[1]))


class ValPubKey(str):
    """`terravalpub-` prefixed Bech32-enconded validator public key, works anywhere a `str`
    is accepted but will perform verification on the checksum.

    :param arg: string-convertable object to convert.
    """

    def __new__(cls, arg):
        validate_val_pubkey(arg)
        return str.__new__(cls, arg)

    @property
    def acc_pubkey(self) -> AccPubKey:
        """The associated account public key."""
        decoded = bech32_decode(self)
        return AccPubKey(bech32_encode("terrapub", decoded[1]))


class ValConsPubKey(str):
    """`terravalpub-` prefixed Bech32-enconded validator consensus public key, works \
    anywhere a `str` is accepted but will perform verification on the checksum.

    :param arg: string-convertable object to convert.
    """

    def __new__(cls, arg):
        validate_val_conspubkey(arg)
        return str.__new__(cls, arg)



#!/usr/bin/env python

import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, "jigu", "__version__.py"), "r") as f:
    exec(f.read(), about)

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

requirements = list()
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author=about["__author__"],
    author_email=about["__author_email__"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description=about["__description__"],
    install_requires=requirements,
    license=about["__license__"],
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="terra jigu sdk blockchain defi finance",
    name=about["__title__"],
    packages=find_packages(include=["jigu", "jigu.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url=about["__url__"],
    version=about["__version__"],
    zip_safe=False,
)



import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.msg.slashing import MsgUnjail
from jigu.error import InvalidValAddress
from jigu.util.validation import is_val_address
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the __schema__ validation set for slashing messages."""
    return load_msg_examples(tdd, msg_types=[MsgUnjail],)


class TestMsgUnjail:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgUnjail]:
            assert_serdes_consistent(MsgUnjail, m)
            assert_serdes_exact(MsgUnjail, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgUnjail.__schema__))
    def test_schema(self, m):
        assert_serdes_consistent(MsgUnjail, m)

    def test_matches_meta(self):
        assert MsgUnjail.type == "cosmos/MsgUnjail"
        assert MsgUnjail.action == "unjail"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, val_address, other):
        """MsgUnjail should validate `address` to be ValAddress.
        """
        assume(not is_val_address(other))
        with pytest.raises(InvalidValAddress):
            MsgUnjail(address=other)
        A = MsgUnjail(address=val_address)
        assert A.address == val_address



import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.msg.staking import (
    MsgBeginRedelegate,
    MsgCreateValidator,
    MsgDelegate,
    MsgEditValidator,
    MsgUndelegate,
)
from jigu.error import InvalidAccAddress, InvalidValAddress
from jigu.util.validation import is_acc_address, is_val_address
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the __schema__ validation set for staking messages."""
    return load_msg_examples(
        tdd,
        msg_types=[
            MsgBeginRedelegate,
            MsgCreateValidator,
            MsgDelegate,
            MsgEditValidator,
            MsgUndelegate,
        ],
    )


class TestMsgBeginRedelegate:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgBeginRedelegate]:
            assert_serdes_consistent(MsgBeginRedelegate, m)
            assert_serdes_exact(MsgBeginRedelegate, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgBeginRedelegate.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgBeginRedelegate, m)

    def test_matches_meta(self):
        assert MsgBeginRedelegate.type == "staking/MsgBeginRedelegate"
        assert MsgBeginRedelegate.action == "begin_redelegate"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, val_address, other):
        """MsgBeginRedelegate should validate `delegator` to be AccAddress, and both
        `validator_src_address` and `validator_dst_address` to be ValAddress.
        """
        assume(not is_acc_address(other) and not is_val_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgBeginRedelegate(
                delegator_address=other,
                validator_src_address=val_address,
                validator_dst_address=val_address,
                amount=None,
            )
        with pytest.raises(InvalidValAddress):
            MsgBeginRedelegate(
                delegator_address=acc_address,
                validator_src_address=other,
                validator_dst_address=val_address,
                amount=None,
            )
        with pytest.raises(InvalidValAddress):
            MsgBeginRedelegate(
                delegator_address=acc_address,
                validator_src_address=val_address,
                validator_dst_address=other,
                amount=None,
            )
        A = MsgBeginRedelegate(
            delegator_address=acc_address,
            validator_src_address=val_address,
            validator_dst_address=val_address,
            amount=None,
        )
        assert A.delegator_address == acc_address
        assert A.validator_src_address == val_address
        assert A.validator_dst_address == val_address


class TestMsgCreateValidator:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgCreateValidator]:
            assert_serdes_consistent(MsgCreateValidator, m)
            assert_serdes_exact(MsgCreateValidator, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgCreateValidator.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgCreateValidator, m)

    def test_matches_meta(self):
        assert MsgCreateValidator.type == "staking/MsgCreateValidator"
        assert MsgCreateValidator.action == "create_validator"


class TestMsgDelegate:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgDelegate]:
            assert_serdes_consistent(MsgDelegate, m)
            assert_serdes_exact(MsgDelegate, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgDelegate.schema))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgDelegate, m)

    def test_matches_meta(self):
        assert MsgDelegate.type == "staking/MsgDelegate"
        assert MsgDelegate.action == "delegate"


class TestMsgEditValidator:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgEditValidator]:
            assert_serdes_consistent(MsgEditValidator, m)
            assert_serdes_exact(MsgEditValidator, m)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgEditValidator.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgEditValidator, m)

    def test_matches_meta(self):
        assert MsgEditValidator.type == "staking/MsgEditValidator"
        assert MsgEditValidator.action == "edit_validator"


class TestMsgUndelegate:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgUndelegate]:
            assert_serdes_consistent(MsgUndelegate, m)
            assert_serdes_exact(MsgUndelegate, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgUndelegate.__schema__))
    def test_schema(self, m):
        assert_serdes_consistent(MsgUndelegate, m)

    def test_matches_meta(self):
        assert MsgUndelegate.type == "staking/MsgUndelegate"
        assert MsgUndelegate.action == "begin_unbonding"



import json

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.proposal import (
    CommunityPoolSpendProposal,
    ParameterChangeProposal,
    RewardWeightUpdateProposal,
    TaxRateUpdateProposal,
    TextProposal,
)
from testtools import assert_serdes_consistent, assert_serdes_exact


class TestTextProposal:
    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(TextProposal.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(TextProposal, m)

    def test_matches_meta(self):
        assert TextProposal.type == "gov/TextProposal"


class TestTaxRateUpdateProposal:
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(TaxRateUpdateProposal.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(TaxRateUpdateProposal, m)

    def test_matches_meta(self):
        assert TaxRateUpdateProposal.type == "treasury/TaxRateUpdateProposal"


class TestRewardWeightUpdateProposal:
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(RewardWeightUpdateProposal.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(RewardWeightUpdateProposal, m)

    def test_matches_meta(self):
        assert RewardWeightUpdateProposal.type == "treasury/RewardWeightUpdateProposal"


class TestCommunityPoolSpendProposal:
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(CommunityPoolSpendProposal.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(CommunityPoolSpendProposal, m)

    def test_matches_meta(self):
        assert (
            CommunityPoolSpendProposal.type == "distribution/CommunityPoolSpendProposal"
        )


class TestParameterChangeProposal:
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(ParameterChangeProposal.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(ParameterChangeProposal, m)

    def test_matches_meta(self):
        assert ParameterChangeProposal.type == "params/ParameterChangeProposal"



from decimal import Decimal

import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core import Dec
from testtools import assert_serdes_consistent, assert_serdes_exact

dec_input = st.one_of(
    st.integers(),
    st.floats(
        allow_infinity=False,
        allow_nan=False,
        min_value=-(10 ** 18),
        max_value=10 ** 18,
    ),
    st.decimals(
        allow_infinity=False,
        allow_nan=False,
        min_value=-(10 ** 25),
        max_value=10 ** 25,
        places=18,
    ),
    from_schema(Dec.__schema__),
)


@pytest.fixture(scope="module")
def inc():
    nd = Dec(0)
    nd.i = 1
    return nd


@pytest.mark.sdk
class TestDecSerdes:
    @pytest.mark.serdes
    @given(d=from_schema(Dec.__schema__))
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    def test_serdes_consistent(self, d):
        assert_serdes_consistent(Dec, d)

    def test_simple_check(self):
        examples = [
            "138875042105.980753034749566779",
            "8447.423744387144096286",
            "3913.113789811986907029",
            "0.500000000000000000",
            "0.006250000000000000",
            "-23.128250000000000023",
            "242.000000000000028422",
            "-242.000000000000020422",
        ]
        for x in examples:
            assert str(Dec(x)) == x

    def test_serializes_zero(self):
        Z = "0.000000000000000000"
        assert str(Dec(0)) == Z
        assert str(Dec("0")) == Z
        assert str(Dec("-0")) == Z
        assert str(Dec("-00")) == Z
        assert str(Dec("-00.0")) == Z
        assert str(Dec("-00.0000000000000000000232")) == Z

    def test_serializes_18_digits(self):
        examples = [
            ("0.5", "0.500000000000000000"),
            ("0.00625", "0.006250000000000000"),
            ("3913.11", "3913.110000000000000000"),
            ("-23.11", "-23.110000000000000000"),
            ("3", "3.000000000000000000"),
            ("-3", "-3.000000000000000000"),
        ]
        for x in examples:
            assert str(Dec(x[0])) == x[1]

    def test_str_regex(self):
        """`Dec` will only parse strings that match Decimal format: `/^(\-)?(\d+)(\.\d+)?\Z/`
        """
        bad_examples = [
            "a1.2",
            "..2",
            "0b00",
            "0xdeadb33f",
            "0.0.5",
            "(626)112-3322",
            "2.14E-14",
            "42.",
            ".420",
        ]
        for bad in bad_examples:
            with pytest.raises(ValueError):
                Dec(bad)

        good_examples = [
            "0.0",
            "42.0",
            "333333.213123923402394",
            "1127.10102392039129301293230921039",
            "13123891823912839123.239482934829348239489",
            "-239299920",
            "-0000.0",
        ]
        for good in good_examples:
            assert isinstance(Dec(good).i, int)

    @given(x=st.integers())
    def test_serializes_int_into_dec(self, x):
        """`Dec` should capture simple `ints` and serialize it to 18 digits
        of precision.
        """
        A = str(x)
        assert str(Dec(A)) == A + "." + 18 * "0"

    @given(
        x=st.decimals(
            allow_infinity=False,
            allow_nan=False,
            min_value=-(10 ** 18),
            max_value=10 ** 18,
            places=18,
        )
    )
    def test_converts_decimal_objects(self, x):
        """`Dec` should capture a compatible `Decimal` object and serialize
        to 18 digits of precision.
        """
        d = Dec(x)
        r = str(d).split(".")
        if x < 0:
            assert d.i < 0
        assert int(r[0]) == int(x)
        assert len(r[1]) == 18

    @given(x=st.integers(), y=st.integers(18, 99))
    def test_truncates_beyond_18_digits(self, x, y):
        """If Dec is provided a string representing a number with higher
        degree of precision than 18 digits, those extra digits should be truncated.
        """
        A = str(x) + "." + y * "7"
        r = str(Dec(A)).split(".")
        assert r[0] == str(x)
        assert len(r[1]) == 18

    @pytest.mark.slow
    @given(x=dec_input)
    def test_constructor_copies(self, x):
        """Dec should copy another Dec in the constructor, and the resulting
        object should be equivalent but not share the same address in memory.
        """
        A = Dec(x)
        B = Dec(A)
        assert A == B
        assert A is not B


@pytest.mark.sdk
class TestDecOperations:
    @pytest.mark.slow
    @given(x=dec_input)
    def test_int(self, x):
        if isinstance(x, str) and "." in x:
            x = x.split(".")[0]
        input_int = int(x)
        output_int = int(Dec(x))
        assert output_int == input_int

    @pytest.mark.slow
    @given(x=dec_input)
    def test_equality_consistent(self, x):
        assert Dec(x) == Dec(x)

    @given(x=st.integers())
    def test_equality_integer(self, x):
        assert Dec(x) == x

    def test_equality_float(self):
        # NOTE: This is about the limit of how robust it is for float
        # In the future, perhaps remove support for float due to lack of utility.
        assert Dec(1.2) == 1.2
        assert Dec(2342.123232) == 2342.123232
        assert Dec(2342.123232123) == 2342.123232123

    def test_equality_decimal(self):
        assert Dec(Decimal("1.45")) == Decimal("1.45")
        assert Dec(Decimal("2342.123232123")) == Decimal("2342.123232123")
        assert Dec(Decimal("2332423442.123234234232123")) == Decimal(
            "2332423442.123234234232123"
        )
        # 24, 18 digit decimal precision
        assert Dec(Decimal("423412312312312342123123.123234234232123123")) == Decimal(
            "423412312312312342123123.123234234232123123"
        )
        # disregard beyond 19 digits of precision
        assert Dec(Decimal("0.1232342342321231233")) == Decimal("0.1232342342321231233")

    @pytest.mark.slow
    @given(x=dec_input, y=dec_input)
    def test_lt_gt(self, x, y):
        assume(not isinstance(x, str) and not isinstance(y, str) and x < y)
        assert Dec(x) < Dec(y)
        assert Dec(y) > Dec(x)
        assert not Dec(x) > Dec(x)
        assert not Dec(x) < Dec(x)
        assert Dec(x) >= Dec(x)
        assert Dec(x) <= Dec(x)
        assert Dec(x) <= Dec(y)
        assert Dec(y) >= Dec(x)

    @pytest.mark.slow
    @given(x=dec_input)
    def test_neg(self, x):
        assume(not isinstance(x, str))
        assert -Dec(x) == Dec(-x)

    @given(x=dec_input)
    def test_abs(self, x):
        assert abs(Dec(x)) >= 0

    @pytest.mark.slow
    @pytest.mark.parametrize("zero", [0, 0.0, Decimal(0), Dec(0)])
    @given(x=dec_input, y=dec_input)
    def test_add_properties(self, x, y, zero):
        assume(x != y)
        assert Dec(x) + 0 == Dec(x)
        assert Dec(x) + 0 == 0 + Dec(x)
        assert Dec(x) + Dec(y) == Dec(y) + Dec(x)
        if not isinstance(x, str) and not isinstance(y, str):
            assert Dec(x) + y == y + Dec(x)
            assert isinstance(Dec(x) + y, Dec)
            assert isinstance(y + Dec(x), Dec)
        if isinstance(x, int) and isinstance(y, int):
            assert Dec(x) + Dec(y) == Dec(x + y)

    def test_add(self, inc):
        assert inc + inc == Dec("0.000000000000000002")
        assert (inc + inc).i == 2

    @pytest.mark.slow
    @pytest.mark.parametrize("zero", [0, 0.0, Decimal(0), Dec(0)])
    @given(x=dec_input, y=dec_input)
    def test_sub_properties(self, x, y, zero):
        assert Dec(x) - zero == Dec(x)
        assert Dec(x) - Dec(y) == -Dec(y) + Dec(x)
        assert Dec(x) - Dec(x) == zero
        assert isinstance(Dec(x) - Dec(y), Dec)
        if not isinstance(y, str):
            assert Dec(x) - y == -y + Dec(x)
            assert isinstance(Dec(x) - y, Dec)
            assert isinstance(y - Dec(x), Dec)
        if isinstance(x, int) and isinstance(y, int):
            assert Dec(x) - Dec(y) == Dec(x - y)

    @pytest.mark.slow
    @pytest.mark.parametrize("one", [1, 1.0, Decimal(1), Dec(1)])
    @given(x=dec_input, y=dec_input)
    def test_mul_properties(self, x, y, one):
        assume(x != y)
        assert Dec(x) * one == Dec(x)
        assert Dec(x) * one == one * Dec(x)
        assert Dec(x) * Dec(y) == Dec(y) * Dec(x)
        if not isinstance(x, str) and not isinstance(y, str):
            assert Dec(x) * y == y * Dec(x)
            assert isinstance(Dec(x) * y, Dec)
            assert isinstance(y * Dec(x), Dec)
        if isinstance(x, int) and isinstance(y, int):
            assert Dec(x) * Dec(y) == Dec(x * y)

    def test_mul_examples(self):
        # TODO: add better examples
        examples = [
            (Dec("6.3") * Dec("6.5"), Dec("40.95")),
            (Dec("27.37") * Dec("36.299"), Dec("993.50363")),
            (Dec("-5231.234232") * 5.5, Dec("-28771.788276")),
        ]
        for x in examples:
            assert x[0] == x[1]

    # TODO: add tests for truediv, floordiv
    def test_cosmos_quo(self):
        examples = [
            (Dec(-1), Dec(-1), Dec(1)),
            (Dec(3), Dec(7), Dec.with_prec(428571428571428571, 18)),
            (
                Dec.with_prec(3333, 4),
                Dec.with_prec(333, 4),
                Dec("10.009009009009009009"),
            ),
        ]
        for x in examples:
            assert x[0] / x[1] == x[2]

    def test_cosmos_mul(self):
        examples = [
            (Dec(-1), Dec(-1), Dec(1)),
            (Dec(3), Dec(7), Dec(21)),
            (Dec.with_prec(3333, 4), Dec.with_prec(333, 4), Dec.with_prec(1109889, 8),),
        ]
        for x in examples:
            assert x[0] * x[1] == x[2]

    def test_cosmos_overflow(self):
        """Analog for this test:
        https://github.com/cosmos/cosmos-sdk/blob/5a2e59ebb23d1d23546f1145d8814a457655935a/types/decimal_test.go#L334
        """
        a = Dec("51643150036226787134389711697696177267")
        b = Dec("-31798496660535729618459429845579852627")
        assert str(a + b) == "19844653375691057515930281852116324640.000000000000000000"



# TODO: Add proper __all__
# __all__ = [
#     "BlocksPerMinute",
#     "BlocksPerHour",
#     "BlocksPerDay",
#     "BlocksPerWeek",
#     "BlocksPerMonth",
#     "BlocksPerYear",
#     "BlocksPerEpoch",
#     "AccAddress",
#     "ValAddress",
#     "",
# ]

BlocksPerMinute = 10
BlocksPerHour = BlocksPerMinute * 60
BlocksPerDay = BlocksPerHour * 24
BlocksPerWeek = BlocksPerDay * 7
BlocksPerMonth = BlocksPerDay * 30
BlocksPerYear = BlocksPerDay * 365
BlocksPerEpoch = BlocksPerWeek

import jigu.core.sdk.address  # isort:skip
from jigu.core.sdk.address import AccAddress, ValAddress  # isort:skip

import jigu.core.sdk.pubkey  # isort:skip
from jigu.core.sdk.pubkey import (  # isort:skip
    AccPubKey,
    PublicKey,
    ValConsPubKey,
    ValPubKey,
)

from .sdk import Coin, Coins, Dec, PublicKey, Timestamp  # isort:skip
from .proposal import Proposal, Content  # isort:skip
from .msg import StdMsg  # isort:skip

from .bank import Input, Output  # isort:skip
from .oracle import ExchangeRatePrevote, ExchangeRateVote  # isort:skip
from .staking import (  # isort:skip
    Commission,
    CommissionRates,
    Delegation,
    Description,
    Redelegation,
    RedelegationEntry,
    UnbondingDelegation,
    UnbondingEntry,
    Validator,
)

from .block import Block  # isort:skip
from .treasury import PolicyConstraints  # isort:skip
from .auth import (  # isort:skip
    Account,
    LazyGradedVestingAccount,
    StdFee,
    StdSignature,
    StdSignMsg,
    StdTx,
    TxBroadcastResult,
    TxInfo,
)
